/*
 * Plot - plot time-based data on screen and to file with interactive controls
 * Copyright (C) 2006  Jonas Berlin <xkr47@outerspace.dyndns.org>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 ****
 * eventhandlers-plot-draw - gtk event handlers related to drawing the
 * plot
 */

#include <math.h>

#include "../system.xh"
#include "eventhandlers-plot-draw.xh"
#include "eventhandlers-plot-mouse.xh"
#include "paintapi-gtk.xh"
#include "plot2-markers.xh"

static void reconfigure(my_plot_view_t *view) {
   view->reconfigured = 0;

   if(view->drawable) g_object_unref(view->drawable);
   view->drawable = gdk_pixmap_new (view->da->window, view->da->allocation.width, view->da->allocation.height, -1);
   paintapi_gtk_set_drawble(view->draw.paintapi, view->drawable);

   plot_size_or_font_changed(view->parent);
}

gboolean plot_configure_event(GtkWidget *widget, GdkEventConfigure *do_not_use, my_plot_view_t *view) {
   if(!view->conc) {
      reconfigure(view);
   } else {
      view->reconfigured++;
   }
   return 0;
}

gboolean plot_map_event(GtkWidget *widget, GdkEvent *do_not_use, my_plot_view_t *view) {
   //printf("  --  map\n");

   view->drawable = gdk_pixmap_new (view->da->window, view->da->allocation.width, view->da->allocation.height, -1);
   view->draw.paintapi = paintapi_gtk_new(view->drawable, &view->draw);
   view->paintapi_direct = paintapi_gtk_new(view->da->window, &view->draw);

   // set cursor
   update_cursor(view);

   // fake configure since it doesn't occur at startup
   plot_configure_event(widget, NULL, view);

   return 0;
}

static void update_fps(my_plot_t *plot) {
   double total_render_time = 0;
   for(int vi=0; vi<plot->view_count; ++vi) {
      total_render_time += plot->view[vi]->draw.last_render_time;
   }

   // TODO smooth the value a bit maybe..
   double fps = total_render_time <= 0 ? INFINITY : 1 / total_render_time;

   char tmp[30];
   if(fps >= 100) {
      sprintf(tmp, "%.0lf", fps);
   } else if(fps >= 10) {
      sprintf(tmp, "%.1lf", fps);
   } else {
      sprintf(tmp, "%.2lf", fps);
   }
   gtk_label_set_text(plot->label_fps, tmp);
}

gboolean plot_expose_event(GtkWidget *widget, GdkEventExpose *event, my_plot_view_t *view) {
   if(view->conc > 0) {
      // still not optimal
      view->normal_expose++;
      //printf(" -> normal_expose++\n");
      return 0;
   }
   int mytrans = ++view->parent->dsc_x->num_changes + view->dsc_y->num_changes;
   view->conc++;
   //    print "->$mytrans view->conc\n";

   //int mvi; for(mvi=0; mvi<view->parent->view_count; ++mvi) if(view->parent->view[mvi] == view) break;
   //printf("%*sexpose %d\n", indent++*3, "", mvi);

   paintapi_gc_t *bggc = view->paintapi_direct->gc_new(view->paintapi_direct);
   paintapi_rgb_t bgcolor = { 0, 0, 0 };
   view->paintapi_direct->gc_set_foreground(view->paintapi_direct, bggc, &bgcolor);

   gdk_draw_rectangle(view->drawable, (GdkGC*)bggc, 1, 0, 0, view->da->allocation.width, view->da->allocation.height);

   if(!view->parent->cursor_marker_gc) {
      view->parent->cursor_marker_gc = gc_new_with_color(&view->draw, view->parent->cursor_marker_color);
      view->draw.paintapi->gc_set_function(view->draw.paintapi, view->parent->cursor_marker_gc, PAINTAPI_FUNCTION_XOR);
   }

   plot_draw(&view->draw);

   // now apply all dynamic stuff to the offscreen buffer => when it is copied to the screen, everything looks like before

   restore_markers_after_draw(view);
   restore_mouse_after_draw(view);

   // copy offscreen buffer to screen

   gdk_draw_drawable(view->da->window, (GdkGC*)bggc, view->drawable, 0, 0, 0, 0, -1, -1);

   view->paintapi_direct->gc_free(view->paintapi_direct, bggc);

   if(view->reconfigured) {
      reconfigure(view);
      view->normal_expose++;
   }

   if(mytrans != view->parent->dsc_x->num_changes + view->dsc_y->num_changes || view->normal_expose) {
      //printf("%d pending exposes, queue draw\n", view->normal_expose);
      //printf("%*sRedraw self (%d)\n", indent*3, "", mvi);
      gtk_widget_queue_draw(view->da);

   } else {
      update_fps(view->parent);

      for(int vi=0; vi<view->parent->view_count; ++vi) {
	 if(view->parent->view[vi]->normal_expose) {
	    //printf("%*sRedraw %d\n", indent*3, "", vi);
	    plot_view_need_redraw(view->parent->view[vi]);
	    break;
	 }
      }
   }

   view->normal_expose = 0;
   view->conc--;

   //printf("%*s/expose %d\n", --indent*3, "", mvi);

   return 0;
}

gboolean plot_focus_event_before(GtkWidget *widget, GdkEventFocus *event, my_plot_view_t *view) {
   // Before focus

   // Save the update area, if any. As a side-effect the current
   // update area is cleared. We gain ownership of the region
   // returned (if any).
   view->update_area = gdk_window_get_update_area(widget->window);

   // return FALSE = continue processing - at some point after this
   // the handlers in GtkWidget will get called, invalidating the
   // whole window unnecessarily.. so we have a second hook after
   // that, see plot_focus_event_after for details.
   return FALSE;
}

gboolean plot_focus_event_after(GtkWidget *widget, GdkEventFocus *event, my_plot_view_t *view) {
   // After focus

   // At this point the GtkWidget handler has invalidated the whole
   // window. We don't want that, so restore the previously
   // invalidated area if any.

   if(view->update_area) {
      // we had an area previously invalidated, restore it
      gdk_window_invalidate_region(widget->window, view->update_area, FALSE);
      gdk_region_destroy(view->update_area);
      view->update_area = NULL;

   } else {
      // Nothing was invalidated before, thus we get the update area
      // (again) which also clears it. We don't need this for
      // anything, so just release it.
      GdkRegion *region = gdk_window_get_update_area(widget->window);
      if(region) {
	 gdk_region_destroy(region);
      }
   }

   return FALSE;
}

void screen_size_changed_event(GdkScreen *screen, my_plot_t *plot) {
   // works only with xrandr, not with ctrl-alt-+ and ctrl-alt--
   plot_size_or_font_changed(plot);
}

/*
 * Local variables:
 * c-file-style: "ellemtel"
 * c-file-offsets: ((c . c-lineup-dont-change) (statement-cont . (lambda (le) (if (save-excursion (goto-char (cdr le)) (looking-at "return")) (c-lineup-java-inher le) (c-lineup-math le)))))
 * End:
 */
