/*
 * Plot - plot time-based data on screen and to file with interactive controls
 * Copyright (C) 2006  Jonas Berlin <xkr47@outerspace.dyndns.org>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 ****
 * eventhandlers-plot-mouse - gtk event handlers related handling
 * mouse events for the plot
 */

#include <gtk/gtktogglebutton.h>
#include <math.h>

#include "eventhandlers-plot-mouse.xh"
#include "eventhandlers-prefs.xh"
#include "plot2-markers.xh"
#include "plot2-zoom-history.xh"

#if 0
#define pprintf printf
#else
#define pprintf(...)
#endif

static void start_hoover_timer(my_plot_view_t *view);
static void stop_hoover_timer(my_plot_view_t *view);
static gboolean hoover_callback(my_plot_view_t *view);
static void remove_hoover(my_plot_view_t *view);

#define BOOKMARK_PROXIMITY (2 * draw->bookmarks.triangle_size)

static int plot_draw_check_if_hit_bookmark(my_plot_draw_t *draw, int x, int y) {
   if(!draw->bookmarks.show || y > BOOKMARK_PROXIMITY) {
      return -1;
   }

   int bi;
   int mbi = -1;
   int mdiff = 0;
   for(bi=0; bi < draw->bookmarks.bookmark_count; ++bi) {
      int diff = abs(x - draw->bookmarks.bookmark[bi].xoffset);
      if(mbi == -1 || diff < mdiff) {
	 mbi = bi;
	 mdiff = diff;
      }
   }

   if(mbi == -1) {
      return -1;
   }

   pprintf("mbi = %d %d %d %lf\n", mbi, mdiff, y, sqrt((double)mdiff * mdiff + y * y));

   if(sqrt((double)mdiff * mdiff + y * y) < BOOKMARK_PROXIMITY) {
      return mbi;
   } else {
      return -1;
   }
}

gboolean plot_button_press_event(GtkWidget *widget, GdkEventButton *event, my_plot_view_t *view) {
   int x = event->x;
   int y = event->y;

   // make sure we receive keyboard events if we are clicked.. dunno
   // why this is needed, maybe because I have this custom handler
   // installed? this function _does_ return 0, which should propagate
   // it further.. :P
   gtk_widget_grab_focus(widget);

   remove_hoover(view);

#ifdef USE_FAM
   boolean_t tailing = gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(view->parent->checkbutton_tail));
#else
   const boolean_t tailing = 0;
#endif

   switch((int)event->type) {
/*
  case GDK_2BUTTON_PRESS: {

  printf("dpress %d %d\n", x, y);
  plot_move_to(view, view->draw.plot_xe/2, view->draw.plot_yde/2);
  break;
  }
*/

      case GDK_BUTTON_PRESS:
	 // printf("press %d %d\n", x, y);
	 switch(view->bmode) {
	    case BMODE_OFF: {

	       view->bx = x;
	       view->by = y;

	       destroy_markers(view->parent);

	       if(y > view->draw.plot_yde) {
		  switch(event->button) {
		     case 1: {
			// place bookmark
			view->bmode = BMODE_MOVE_BOOKMARK;
			view->bbookmark_is_new = true;
			view->bbookmark_xoff = 0;

			int id = view->draw.bookmarks.bookmark_count;
			my_plot_bookmark_arr_set_length(view->draw.bookmarks.bookmark, id + 1);

			view->bbookmark_id = id;

			plot_motion_generic(view, x, y, 0);
			break;
		     }
		  }

		  return 0;
	       }

	       if(plot_draw_check_if_hit_legend_box(&view->draw, x, y)) {
		  int lsi, li;
		  my_plot_line_info_t *line = plot_draw_check_if_hit_legend_item(&view->draw, x, y, &lsi, &li);

		  if(line) {
		     switch(event->button) {
			case 1:
			   view->bstate = line->enabled = !line->enabled;
			   view->bmode = BMODE_LEGEND_ENABLE_DISABLE;

			   GtkTreePath *path = gtk_tree_path_new_from_indices(lsi, li, -1);
			   prefs_line_enabled_changed(view->parent, path);

			   plot_view_need_redraw(view);
			   update_view_minmax(view);
			   break;

			   /*
			  default:
			  diff mode, TODO scales don't work and we need a time unit also
			  view->bstate = line->mode == LINE_MODE_NORMAL;
			  line->mode = view->bstate ? LINE_MODE_DIFF : LINE_MODE_NORMAL;
			  view->bmode = BMODE_LEGEND_DIFF_NORMAL;

			  plot_view_need_redraw(view);
			  update_view_minmax(view);
			  break;
			*/
		     }
		  }
		  return 0;
	       }

	       // maybe we want to move a bookmark ?

	       int mbi = plot_draw_check_if_hit_bookmark(&view->draw, x, y);
	       if(mbi != -1) {
		  switch(event->button) {
		     case 1: {
			// move bookmark
			view->bmode = BMODE_MOVE_BOOKMARK;
			view->bbookmark_is_new = false;
			view->bbookmark_xoff = x - view->draw.bookmarks.bookmark[mbi].xoffset;
			view->bbookmark_id = mbi;

			plot_motion_generic(view, x, y, 0);
			break;
		     }
		  }

		  return 0;
	       }

	       // update scale pos according to press location

	       int snap_x_threshold = 40; // view->draw.plot_xe / 10;
	       int snap_y_threshold = 40; // view->draw.plot_yde / 10;
	       int snap_x = x < snap_x_threshold ? 0 : x >= view->draw.plot_xe - snap_x_threshold ? view->draw.plot_xe-1 : x;
	       int snap_y = y < snap_y_threshold ? 0 : y >= view->draw.plot_yde - snap_y_threshold ? view->draw.plot_yde-1 : y;

	       double new_x_zoompoint = ((double)snap_x) / view->draw.plot_xe;
	       double new_y_zoompoint = 1 - ((double)snap_y) / view->draw.plot_yde;

	       view->bspeed = event->state & GDK_SHIFT_MASK ? 4 : 1;

	       if(event->button > 1) {
		  view->bmode = BMODE_INTERACTIVE_ZOOM;
		  view->bxloglength = data_scrollbar_controller_get_page_length_log(view->parent->dsc_x);
		  view->byloglength = data_scrollbar_controller_get_page_length_log(view->dsc_y);

		  // when close to border, zoom to border
		  if(tailing || x >= view->draw.plot_xe - 30) {
		     new_x_zoompoint = 1.0;
		  } else if(x < 30) { // ##TODO## conf 30
		     new_x_zoompoint = 0.0;
		  }

		  if(y < 30) {
		     new_y_zoompoint = 1.0;
		  } else if(y >= view->draw.plot_yde - 30) {
		     new_y_zoompoint = 0.0;
		  }

	       } else if(event->state & GDK_CONTROL_MASK) {
		  view->bmode = BMODE_ZOOM_AREA;
		  if(tailing) {
		     view->bx = view->draw.plot_xe;
		  }
		  view->box = COORD_UNDEF;

	       } else {
		  if(tailing) {
		     new_x_zoompoint = 1.0;
		     view->bx = view->draw.plot_xe;
		  }
		  view->bmode = BMODE_INTERACTIVE_MOVE;
		  view->bxoff = data_scrollbar_controller_get_page_off(view->parent->dsc_x);
		  view->byoff = data_scrollbar_controller_get_page_off(view->dsc_y);
	       }

	       data_scrollbar_controller_set_zoompoint(view->parent->dsc_x, new_x_zoompoint);
	       data_scrollbar_controller_set_zoompoint(view->dsc_y, new_y_zoompoint);

	       update_cursor(view);
	       break;
	    }

	    case BMODE_MOVE_BOOKMARK: {
	       if(event->button > 1) {

		  view->bx = x;
		  view->by = y;

		  // ---

		  int snap_x_threshold = 40; // view->draw.plot_xe / 10;
		  int snap_y_threshold = 40; // view->draw.plot_yde / 10;
		  int snap_x = x < snap_x_threshold ? 0 : x >= view->draw.plot_xe - snap_x_threshold ? view->draw.plot_xe-1 : x;
		  int snap_y = y < snap_y_threshold ? 0 : y >= view->draw.plot_yde - snap_y_threshold ? view->draw.plot_yde-1 : y;

		  double new_x_zoompoint = ((double)snap_x) / view->draw.plot_xe;
		  double new_y_zoompoint = 1 - ((double)snap_y) / view->draw.plot_yde;

		  view->bspeed = event->state & GDK_SHIFT_MASK ? 4 : 1;

		  // ---

		  view->bmode = BMODE_MOVE_BOOKMARK_INTERACTIVE_ZOOM;
		  view->bxloglength = data_scrollbar_controller_get_page_length_log(view->parent->dsc_x);
		  view->byloglength = data_scrollbar_controller_get_page_length_log(view->dsc_y);

		  // when close to border, zoom to border
		  if(tailing || x >= view->draw.plot_xe - 30) {
		     new_x_zoompoint = 1.0;
		  } else if(x < 30) { // ##TODO## conf 30
		     new_x_zoompoint = 0.0;
		  }

		  if(y < 30) {
		     new_y_zoompoint = 1.0;
		  } else if(y >= view->draw.plot_yde - 30) {
		     new_y_zoompoint = 0.0;
		  }

		  data_scrollbar_controller_set_zoompoint(view->parent->dsc_x, new_x_zoompoint);
		  data_scrollbar_controller_set_zoompoint(view->dsc_y, new_y_zoompoint);

		  update_cursor(view);
	       }
	       break;
	    }

	    default:
	       break;
	 }
	 break;
   }

   return 0;
}

#define LIMIT(v,min,max) do{ if(v<(min)) v=min; else if(v>(max)) v=max; } while(0)

#define HOOVER_TIMEOUT 200

static void start_hoover_timer(my_plot_view_t *view) {
   if(view->hoover_timer_tag) {
      fprintf(stderr, "WARNING: previous hoover timer not stopped\n");
      stop_hoover_timer(view);
   }

   view->hoover_timer_tag = g_timeout_add(HOOVER_TIMEOUT, (GSourceFunc)hoover_callback, view);
}

static void stop_hoover_timer(my_plot_view_t *view) {
   if(view->hoover_timer_tag) {
      g_source_remove(view->hoover_timer_tag);
      view->hoover_timer_tag = 0;
   }
}

static gboolean hoover_callback(my_plot_view_t *view) {
   view->hoover_timer_tag = 0;

   if(view->by > view->draw.plot_yde) return 0;

   view->hoover_painted = 1;

   double ox = data_scrollbar_controller_get_page_off(view->parent->dsc_x);
   double oy = data_scrollbar_controller_get_page_off(view->dsc_y);
   double lx = data_scrollbar_controller_get_page_length(view->parent->dsc_x);
   double ly = data_scrollbar_controller_get_page_length(view->dsc_y);

   int y = view->draw.plot_yde - view->by;

   ox += lx * view->bx / view->draw.plot_xe;
   oy += ly *        y / view->draw.plot_yde;

/*
   lx *= (x - view->bx) / (double)view->draw.plot_xe;
   ly *= (y - view->by) / (double)view->draw.plot_yde;
*/
   pprintf("Timeout at %d,%d -> x = %lf (%s)  y = %lf\n", view->bx, view->by, ox / TIME_MULTIPLIER, format_time((my_time_t)ox, 1), oy);

   return 0;
}

// TODO what to do with hoover before/while/after redrawing plot

static void remove_hoover(my_plot_view_t *view) {
   stop_hoover_timer(view);

   if(!view->hoover_painted) return;
   view->hoover_painted = 0;

   pprintf("Hoover removed\n");
}

// return true if release accepted.. return value doesn't matter if release_button == 0
boolean_t plot_motion_generic(my_plot_view_t *view, int x, int y, int release_button) {
#ifdef USE_FAM
   boolean_t tailing = gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(view->parent->checkbutton_tail));
#else
   const boolean_t tailing = 0;
#endif

   pprintf("mode = %d release = %d\n", view->bmode, release_button);

   switch(view->bmode) {
      case BMODE_OFF:
	 view->bx = x;
	 view->by = y;

	 if(!gtk_widget_is_focus(view->da)) {
	    gtk_widget_grab_focus(view->da);
	 }

	 remove_hoover(view);

	 view->bbookmark_id = plot_draw_check_if_hit_bookmark(&view->draw, x, y);

	 if((view->inside_legend = plot_draw_check_if_hit_legend_box(&view->draw, x, y))) {
	    destroy_markers(view->parent);
	 } else {
	    set_and_show_markers(view, x, y);
	    start_hoover_timer(view);
	 }
	 update_cursor(view);
	 break;

      case BMODE_INTERACTIVE_MOVE:
	 plot_move_to(view, tailing ? view->draw.plot_xe : x, y);
	 if(release_button) {
	    add_to_zoom_history(view->parent);
	 }
	 break;

      case BMODE_INTERACTIVE_ZOOM:
      case BMODE_MOVE_BOOKMARK_INTERACTIVE_ZOOM: {
	 ++view->parent->disable_delayed_add_zoom;

	 // ##TODO## conf 5.0 and 2.0
	 data_scrollbar_controller_set_page_length_log(view->parent->dsc_x, view->bxloglength - (x - view->bx) * view->bspeed * 5.0 / view->draw.plot_xe);
	 data_scrollbar_controller_set_page_length_log(view->dsc_y, view->byloglength - (y - view->by) * view->bspeed * 2.0 / view->draw.plot_yde);

	 --view->parent->disable_delayed_add_zoom;

	 switch(view->bmode) {
	    case BMODE_MOVE_BOOKMARK_INTERACTIVE_ZOOM:
	       switch(release_button) {
		  case 0:
		     // no release, do nothing
		     break;
		  case 1:
		     // will fall through below and end the zooming + bookmark dragging
		     break;
		  default:
		     view->bmode = BMODE_MOVE_BOOKMARK;
		     update_cursor(view);
		     return 0;
	       }
	       break;

	    default:
	       break;
	 }

	 if(release_button) {
	    add_to_zoom_history(view->parent);
	 }

	 if(view->bmode != BMODE_MOVE_BOOKMARK_INTERACTIVE_ZOOM)
	    break;
	 // else fall through
      }

      case BMODE_MOVE_BOOKMARK: {
	 if(release_button == 1) return 1;

	 x -= view->bbookmark_xoff;

	 LIMIT(x, 0, view->draw.plot_xe);

	 my_plot_bookmark_t *bm = &view->draw.bookmarks.bookmark[view->bbookmark_id];

	 if(bm->bookmark) free(bm->bookmark);

	 double ox = data_scrollbar_controller_get_page_off(view->parent->dsc_x);
	 double lx = data_scrollbar_controller_get_page_length(view->parent->dsc_x);

	 my_time_t time = (my_time_t)round(ox + lx * x / view->draw.plot_xe);

	 bm->timestamp = time;
	 bm->bookmark = g_strdup_printf(release_button == 1 ? "%s" : view->bbookmark_is_new ? "new: %s" : "move: %s", format_time(bm->timestamp, 1));

	 plot_view_need_redraw(view);

	 return 0;
      }

      case BMODE_ZOOM_AREA: {
	 LIMIT(view->bx, 0, view->draw.plot_xe);
	 LIMIT(view->by, 0, view->draw.plot_yde);
	 LIMIT(x, 0, view->draw.plot_xe);
	 LIMIT(y, 0, view->draw.plot_yde);

	 view->paintapi_direct->gc_set_dashes(view->paintapi_direct, view->parent->cursor_marker_gc, -view->hline, cursor_marker_dashes, SZ(cursor_marker_dashes));
	 if(!release_button) {
	    view->paintapi_direct->draw_rectangle(view->paintapi_direct, view->parent->cursor_marker_gc, 0, view->bx, view->by, x, y);
	 }
	 if(view->box != COORD_UNDEF) {
	    view->paintapi_direct->draw_rectangle(view->paintapi_direct, view->parent->cursor_marker_gc, 0, view->bx, view->by, view->box, view->boy);
	 }
	 if(!release_button) {
	    view->box = x;
	    view->boy = y;
	 } else {
	    int x1 = view->bx;
	    int y1 = view->by;

	    y = view->draw.plot_yde - y;
	    y1 = view->draw.plot_yde - y1;

	    if(x < x1) { int t = x; x = x1; x1 = t; }
	    if(y < y1) { int t = y; y = y1; y1 = t; }

	    double ox = data_scrollbar_controller_get_page_off(view->parent->dsc_x);
	    double oy = data_scrollbar_controller_get_page_off(view->dsc_y);
	    double lx = data_scrollbar_controller_get_page_length(view->parent->dsc_x);
	    double ly = data_scrollbar_controller_get_page_length(view->dsc_y);

	    //printf("%d,%d %d,%d\n", x1, y1, x-x1, y-y1);

	    //printf("%lf,%lf %lf,%lf\n", ox, oy, lx, ly);

	    ox += lx * x1 / view->draw.plot_xe;
	    oy += ly * y1 / view->draw.plot_yde;

	    lx *= (x - x1) / (double)view->draw.plot_xe;
	    ly *= (y - y1) / (double)view->draw.plot_yde;

	    //printf("%lf,%lf %lf,%lf\n\n", ox, oy, lx, ly);

	    ++view->parent->disable_delayed_add_zoom;

	    // after area zoom, always put center as zoom point
	    data_scrollbar_controller_set_zoompoint(view->parent->dsc_x, 0.5);
	    data_scrollbar_controller_set_zoompoint(view->dsc_y, 0.5);

	    // update page lengths
	    data_scrollbar_controller_set_page_length(view->parent->dsc_x, lx);
	    data_scrollbar_controller_set_page_length(view->dsc_y, ly);

	    // now, in case either page length subceeded the minimum page length, we adjust offset so that the selected zoom area becomes centered in view
	    double nlx = data_scrollbar_controller_get_page_length(view->parent->dsc_x);
	    double nly = data_scrollbar_controller_get_page_length(view->dsc_y);

	    ox -= (nlx - lx) / 2;
	    oy -= (nly - ly) / 2;

	    // update offsets
	    data_scrollbar_controller_set_page_off(view->parent->dsc_x, ox);
	    data_scrollbar_controller_set_page_off(view->dsc_y, oy);

	    //printf("%lf,%lf %lf,%lf\n\n", data_scrollbar_controller_get_page_off(view->parent->dsc_x), data_scrollbar_controller_get_page_off(view->dsc_y), nlx, nly);

	    view->box = COORD_UNDEF;

	    --view->parent->disable_delayed_add_zoom;

	    add_to_zoom_history(view->parent);
	 }
	 break;
      }

      case BMODE_LEGEND_ENABLE_DISABLE: {
	 if(!view->draw.show_disabled) break;
	 if(!plot_draw_check_if_hit_legend_box(&view->draw, x, y)) break;

	 int lsi, li;
	 my_plot_line_info_t *line = plot_draw_check_if_hit_legend_item(&view->draw, x, y, &lsi, &li);

	 if(line) {
	    int changed = line->enabled != view->bstate;
	    if(changed) {
	       line->enabled = view->bstate;

	       GtkTreePath *path = gtk_tree_path_new_from_indices(lsi, li, -1);
	       prefs_line_enabled_changed(view->parent, path);

	       plot_view_need_redraw(view);
	       update_view_minmax(view);
	    }
	 }
	 break;
      }

/*
      case BMODE_LEGEND_DIFF_NORMAL: {
	 if(!plot_draw_check_if_hit_legend_box(&view->draw, x, y)) break;

	 my_plot_line_info_t *line = plot_draw_check_if_hit_legend_item(&view->draw, x, y);

	 if(line) {
	    int changed = line->mode != (view->bstate ? LINE_MODE_DIFF : LINE_MODE_NORMAL);
	    if(changed) {
	       line->mode = view->bstate ? LINE_MODE_DIFF : LINE_MODE_NORMAL;

	       plot_view_need_redraw(view);
	       update_view_minmax(view);
	    }
	 }
	 break;
      }
*/
   }

   return 1;
}

void restore_mouse_after_draw(my_plot_view_t *view) {
   if(view->box != COORD_UNDEF) {
      view->draw.paintapi->gc_set_dashes(view->draw.paintapi, view->parent->cursor_marker_gc, -view->hline, cursor_marker_dashes, SZ(cursor_marker_dashes));
      view->draw.paintapi->draw_rectangle(view->draw.paintapi, view->parent->cursor_marker_gc, 0, view->bx, view->by, view->box, view->boy);
   }
}

gboolean plot_motion_notify_event(GtkWidget *wid, GdkEventMotion *event, my_plot_view_t *view) {
   int x = event->x;
   int y = event->y;

   plot_motion_generic(view, x, y, 0);

   return 0;
}

gboolean plot_button_release_event(GtkWidget *widget, GdkEventButton *event, my_plot_view_t *view) {
   int x = event->x;
   int y = event->y;

   if(plot_motion_generic(view, x, y, (int)event->button)) {

      view->bmode = BMODE_OFF;

      // call generic motion routine again, this time in BMODE_OFF and as
      // if the pointer just moved to the same location which lets the
      // routine set up anything that's usually established when no
      // button is pressed and the mouse is moved.
      plot_motion_generic(view, x, y, 0);

      update_cursor(view);
   }

   return 0;
}

gboolean plot_enter_notify_event(GtkWidget *widget, GdkEventCrossing *event, my_plot_view_t *view) {
   int x = event->x;
   int y = event->y;

   plot_motion_generic(view, x, y, 0);

   return 0;
}

gboolean plot_leave_notify_event(GtkWidget *widget, GdkEventCrossing *event, my_plot_view_t *view) {
   if(view->bmode == BMODE_OFF) {
      destroy_markers(view->parent);
      stop_hoover_timer(view);
   }

   return 0;
}

/*
 * Local variables:
 * c-file-style: "ellemtel"
 * c-file-offsets: ((c . c-lineup-dont-change) (statement-cont . (lambda (le) (if (save-excursion (goto-char (cdr le)) (looking-at "return")) (c-lineup-java-inher le) (c-lineup-math le)))))
 * End:
 */
