/*
 * Plot - plot time-based data on screen and to file with interactive controls
 * Copyright (C) 2006  Jonas Berlin <xkr47@outerspace.dyndns.org>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 ****
 * eventhandlers-prefs - gtk event handlers for the prefs TreeView
 */

#include <gtk/gtkcellrenderertext.h>
#include <gtk/gtkcellrenderertoggle.h>
#include <gtk/gtkcolorsel.h>
#include <gtk/gtkdialog.h>
#include <gtk/gtktogglebutton.h>
#include <gtk/gtktreeselection.h>
#include <gtk/gtktreeview.h>

#include "eventhandlers-prefs.xh"

static gboolean prefs_search_equal(GtkTreeModel *model, gint column, const gchar *key, GtkTreeIter *iter, my_plot_t *plot) {
   if(column != 0) {
      return false;
   }
   GtkTreeView *prefs = GTK_TREE_VIEW(plot->treeview_prefs);
   GtkTreeStore *store = GTK_TREE_STORE(gtk_tree_view_get_model(prefs));
   gchar *str;
   gtk_tree_model_get(GTK_TREE_MODEL(store), iter, 0, &str, -1);
   gboolean found = strstr(str, key) != NULL;
   free(str);
   return !found;
}

static void prefs_name_edited(GtkCellRendererText *renderer, gchar *pathstr, gchar *newtext, my_plot_t *plot) {
   GtkTreePath *path = gtk_tree_path_new_from_string(pathstr);
   if(!path) {
      fprintf(stderr, "failed to create GtkTreePath for %s\n", pathstr);
      return;
   }
   GtkTreeView *prefs = GTK_TREE_VIEW(plot->treeview_prefs);
   GtkTreeStore *store = GTK_TREE_STORE(gtk_tree_view_get_model(prefs));
   int depth = gtk_tree_path_get_depth(path);
   switch(depth) {
      case 1: {
	 int *indexes = gtk_tree_path_get_indices(path);
	 int lsi = indexes[0];
	 my_dataset_t *dataset = plot->view[0]->draw.line_set[lsi].dataset;

	 free(dataset->name);
	 dataset->name = strdup(newtext);
	 break;
      }
      case 2: {
	 int *indexes = gtk_tree_path_get_indices(path);
	 int lsi = indexes[0];
	 int li = indexes[1];
	 my_plot_line_info_set_t *line_set = &plot->view[0]->draw.line_set[lsi];
	 my_plot_line_info_t *line = &line_set->line[li];
	 my_dataset_t *dataset = line_set->dataset;

	 free(dataset->item[line->dataset_idx].name);
	 dataset->item[line->dataset_idx].name = strdup(newtext);
	 break;
      }
      default:
	 fprintf(stderr, "bad tree path depth for %s\n", pathstr);
	 break;
   }

   GtkTreeIter iter;
   if(gtk_tree_model_get_iter(GTK_TREE_MODEL(store), &iter, path)) {
      gtk_tree_store_set(store, &iter, 0, newtext, -1);
   }

   gtk_tree_path_free(path);

   plot_need_redraw(plot);
}

static void prefs_line_set_enabled_changed(my_plot_t *plot, GtkTreePath *path) {
   int *indexes = gtk_tree_path_get_indices(path);
   int lsi = indexes[0];
   gboolean all_enabled = true;
   gboolean all_disabled = true;
   for(int vi=0; vi<plot->view_count; ++vi) {
      for(int li=0; li<plot->view[vi]->draw.line_set[lsi].line_count; ++li) {
	 gboolean val = plot->view[vi]->draw.line_set[lsi].line[li].enabled;
	 if(!val) all_enabled = false;
	 else all_disabled = false;
      }
   }

   GtkTreeView *prefs = GTK_TREE_VIEW(plot->treeview_prefs);
   GtkTreeStore *store = GTK_TREE_STORE(gtk_tree_view_get_model(prefs));
   GtkTreeIter iter;
   if(gtk_tree_model_get_iter(GTK_TREE_MODEL(store), &iter, path)) {
      gtk_tree_store_set(store, &iter, 1, all_enabled, 2, !all_enabled && !all_disabled, -1);
   }
   gtk_tree_path_free(path);
}

void prefs_line_enabled_changed(my_plot_t *plot, GtkTreePath *path) {
   int *indexes = gtk_tree_path_get_indices(path);
   int lsi = indexes[0];
   int li = indexes[1];
   gboolean all_enabled = true;
   gboolean all_disabled = true;
   for(int vi=0; vi<plot->view_count; ++vi) {
      gboolean val = plot->view[vi]->draw.line_set[lsi].line[li].enabled;
      if(!val) all_enabled = false;
      else all_disabled = false;
   }

   GtkTreeView *prefs = GTK_TREE_VIEW(plot->treeview_prefs);
   GtkTreeStore *store = GTK_TREE_STORE(gtk_tree_view_get_model(prefs));
   GtkTreeIter iter;
   if(gtk_tree_model_get_iter(GTK_TREE_MODEL(store), &iter, path)) {
      gtk_tree_store_set(store, &iter, 1, all_enabled, 2, !all_enabled && !all_disabled, -1);
   }
   gtk_tree_path_free(path);

   GtkTreePath *parent_path = gtk_tree_path_new_from_indices(lsi, -1);
   prefs_line_set_enabled_changed(plot, parent_path);
}

static void prefs_enabled_toggled(GtkCellRendererToggle *cell_renderer, gchar *pathstr, my_plot_t *plot) {
   GtkTreePath *path = gtk_tree_path_new_from_string(pathstr);
   if(!path) {
      fprintf(stderr, "failed to create GtkTreePath for %s\n", pathstr);
      return;
   }
   GtkTreeView *prefs = GTK_TREE_VIEW(plot->treeview_prefs);
   GtkTreeStore *store = GTK_TREE_STORE(gtk_tree_view_get_model(prefs));
   int depth = gtk_tree_path_get_depth(path);
   switch(depth) {
      case 1: {
	 int *indexes = gtk_tree_path_get_indices(path);
	 int lsi = indexes[0];
	 gboolean all_enabled = true;
	 for(int li=0; li<plot->view[0]->draw.line_set[lsi].line_count; ++li) {
	    gboolean val = plot->view[0]->draw.line_set[lsi].line[li].enabled;
	    if(!val) {
	       all_enabled = false;
	       break;
	    }
	 }
	 gboolean val = !all_enabled;
	 for(int vi=0; vi<plot->view_count; ++vi) {
	    for(int li=0; li<plot->view[vi]->draw.line_set[lsi].line_count; ++li) {
	       plot->view[vi]->draw.line_set[lsi].line[li].enabled = val;
	    }
	 }

	 GtkTreeIter child;
	 GtkTreeIter iter;
	 if(!gtk_tree_model_get_iter(GTK_TREE_MODEL(store), &iter, path)) break;
	 if(!gtk_tree_model_iter_children(GTK_TREE_MODEL(store), &child, &iter)) break;
	 do {
	    gtk_tree_store_set(store, &child, 1, val, 2, false, -1);
	 } while(gtk_tree_model_iter_next(GTK_TREE_MODEL(store), &child));
	 prefs_line_set_enabled_changed(plot, path);
	 break;
      }
      case 2: {
	 int *indexes = gtk_tree_path_get_indices(path);
	 int lsi = indexes[0];
	 int li = indexes[1];
	 gboolean val = !plot->view[0]->draw.line_set[lsi].line[li].enabled;
	 for(int vi=0; vi<plot->view_count; ++vi) {
	    plot->view[vi]->draw.line_set[lsi].line[li].enabled = val;
	 }
	 prefs_line_enabled_changed(plot, path);
	 break;
      }
      default:
	 fprintf(stderr, "bad tree path depth for %s\n", pathstr);
	 gtk_tree_path_free(path);
	 break;
   }

   plot_need_redraw(plot);
}


// popup menu

static void prefs_line_properties_dialog(my_plot_t *plot, int lsi, int li) {
   GladeXML *gladexml = glade_xml_new(plot->gladefile, "dialog_line_properties", NULL);
   GtkWidget *dialog = glade_xml_get_widget(gladexml, "dialog_line_properties");

   GtkToggleButton *line_type_normal = GTK_TOGGLE_BUTTON(glade_xml_get_widget(gladexml, "radiobutton_line_type_normal"));
   GtkToggleButton *line_type_step = GTK_TOGGLE_BUTTON(glade_xml_get_widget(gladexml, "radiobutton_line_type_step"));
   GtkColorSelection *line_color = GTK_COLOR_SELECTION(glade_xml_get_widget(gladexml, "colorselection_line_color"));

   my_plot_line_info_t *line = &plot->view[0]->draw.line_set[lsi].line[li];

   GdkColor col = {
      .red = line->color.r,
      .green = line->color.g,
      .blue = line->color.b,
   };

   gtk_color_selection_set_has_opacity_control(line_color, false);
   gtk_color_selection_set_current_color(line_color, &col);
   gtk_color_selection_set_previous_color(line_color, &col);
   gtk_color_selection_set_has_palette(line_color, true);

   gtk_toggle_button_set_active(line->linetype == TYPE_LINE ? line_type_normal : line_type_step, true);

   while(42) {
      int res = gtk_dialog_run(GTK_DIALOG(dialog));

      if(res == GTK_RESPONSE_OK || res == GTK_RESPONSE_APPLY) {
	 if(res == GTK_RESPONSE_OK) {
	    gtk_widget_hide(dialog);
	 }
	 gtk_color_selection_get_current_color(line_color, &col);
	 line->color.r = col.red;
	 line->color.g = col.green;
	 line->color.b = col.blue;
	 line->linetype = gtk_toggle_button_get_active(line_type_step) ? TYPE_STEP : TYPE_LINE;

	 for(int vi=0; vi<plot->view_count; ++vi) {
	    plot_draw_reset_line_info_gcs(&plot->view[vi]->draw, line);
	 }
	 plot_need_redraw(plot);
      }

      if(res != GTK_RESPONSE_APPLY) {
	 break;
      }
   }

   gtk_widget_destroy(dialog);
}

static void prefs_line_properties_clicked(GtkWidget *menuitem, my_plot_t *plot) {
   GtkTreeSelection *selection;
   selection = gtk_tree_view_get_selection(GTK_TREE_VIEW(plot->treeview_prefs));

   GList *selected = gtk_tree_selection_get_selected_rows(selection, NULL);

   GtkTreePath *first = (GtkTreePath*)selected->data;
   int *indexes = gtk_tree_path_get_indices(first);
   prefs_line_properties_dialog(plot, indexes[0], indexes[1]);

   g_list_foreach(selected, (GFunc)gtk_tree_path_free, NULL);
   g_list_free(selected);
}

static gboolean prefs_show_popup(GdkEventButton *event, my_plot_t *plot) {
   /* make sure a line is selected - line sets don't have properties */
   GtkTreeSelection *selection;
   selection = gtk_tree_view_get_selection(GTK_TREE_VIEW(plot->treeview_prefs));

   GList *selected = gtk_tree_selection_get_selected_rows(selection, NULL);

   GtkTreePath *first = (GtkTreePath*)selected->data;
   int depth = gtk_tree_path_get_depth(first);
   g_list_foreach(selected, (GFunc)gtk_tree_path_free, NULL);
   g_list_free(selected);

   gtk_widget_set_sensitive(plot->menuitem_line_properties, depth == 2);

   gtk_menu_popup(GTK_MENU(plot->popupmenu_line), NULL, NULL, NULL, NULL,
		  (event != NULL) ? event->button : 0,
		  gdk_event_get_time((GdkEvent*)event));
   return TRUE;
}

static gboolean prefs_clicked(GtkWidget *treeview, GdkEventButton *event, my_plot_t *plot) {
   if (event->type != GDK_BUTTON_PRESS || event->button != 3) {
      return FALSE;
   }

   GtkTreeSelection *selection;
   selection = gtk_tree_view_get_selection(GTK_TREE_VIEW(treeview));

   if(gtk_tree_selection_count_selected_rows(selection) <= 1) {
      GtkTreePath *path;
      if(gtk_tree_view_get_path_at_pos(GTK_TREE_VIEW(treeview), (int)event->x, (int)event->y, &path, NULL, NULL, NULL)) {
	 gtk_tree_selection_unselect_all(selection);
	 gtk_tree_selection_select_path(selection, path);
	 gtk_tree_path_free(path);
      }
      if(gtk_tree_selection_count_selected_rows(selection) < 1) {
	 return FALSE;
      }
   }
   return prefs_show_popup(event, plot);
}

static gboolean prefs_popup(GtkWidget *treeview, my_plot_t *plot) {
   return prefs_show_popup(NULL, plot);
}


// init

void prefs_init(my_plot_t *plot) {
   GtkTreeView *prefs = GTK_TREE_VIEW(plot->treeview_prefs);

   gtk_tree_view_set_headers_visible(prefs, true);

   gtk_tree_view_set_rules_hint(prefs, true);

   //gtk_tree_view_set_reorderable(prefs, true);
   // TODO do it manually so we can control what operations are allowed - check what set_reorderable() does
   // TODO listen to row_inserted and row_deleted to get info about reordering..

   gtk_tree_view_set_enable_search(prefs, true);
   gtk_tree_view_set_search_column(prefs, 0);
   gtk_tree_view_set_search_equal_func(prefs, prefs_search_equal, plot, NULL);

   // Only enable this option if all rows are the same height and all columns are of type GTK_TREE_VIEW_COLUMN_FIXED.
   gtk_tree_view_set_fixed_height_mode(prefs, false);

   // model

   GType types[] = {
      [PREFS_COLUMN_NAME] = G_TYPE_STRING,
      [PREFS_COLUMN_ENABLED] = G_TYPE_BOOLEAN,
      [PREFS_COLUMN_ENABLED_PARTIALLY] = G_TYPE_BOOLEAN,
      [PREFS_COLUMN_MIN] = G_TYPE_STRING,
      [PREFS_COLUMN_MAX] = G_TYPE_STRING,
      [PREFS_COLUMN_OTHER] = G_TYPE_STRING,
   };

   GtkTreeStore *store = gtk_tree_store_newv(PREFS_N_COLUMNS, types);

   // setup columns

   GtkCellRenderer *renderer;
   GtkTreeViewColumn *column;

   renderer = gtk_cell_renderer_text_new();
   g_object_set(G_OBJECT(renderer), "editable", true, NULL);
   g_object_set(G_OBJECT(renderer), "editable-set", true, NULL);
   g_signal_connect(renderer, "edited", G_CALLBACK(prefs_name_edited), plot);
   column = gtk_tree_view_column_new_with_attributes("Name", renderer, "text", PREFS_COLUMN_NAME, NULL);
   gtk_tree_view_append_column(prefs, column);

   renderer = gtk_cell_renderer_toggle_new();
   g_signal_connect(renderer, "toggled", G_CALLBACK(prefs_enabled_toggled), plot);
   column = gtk_tree_view_column_new_with_attributes("Enabled", renderer,
						     "active", PREFS_COLUMN_ENABLED,
						     "inconsistent", PREFS_COLUMN_ENABLED_PARTIALLY,
						     NULL);
   gtk_tree_view_append_column(prefs, column);
   renderer = gtk_cell_renderer_text_new();

   column = gtk_tree_view_column_new_with_attributes("Min", renderer, "text", PREFS_COLUMN_MIN, NULL);
   gtk_tree_view_append_column(prefs, column);

   renderer = gtk_cell_renderer_text_new();
   column = gtk_tree_view_column_new_with_attributes("Max", renderer, "text", PREFS_COLUMN_MAX, NULL);
   gtk_tree_view_append_column(prefs, column);

   renderer = gtk_cell_renderer_text_new();
   column = gtk_tree_view_column_new_with_attributes("Other", renderer, "text", PREFS_COLUMN_OTHER, NULL);
   gtk_tree_view_append_column(prefs, column);

   gtk_tree_view_set_model(prefs, GTK_TREE_MODEL(store));

   // popup menus

   GladeXML *gladexml_popup = glade_xml_new(plot->gladefile, "popupmenu_line", NULL);
   plot->popupmenu_line = glade_xml_get_widget(gladexml_popup, "popupmenu_line");
   plot->menuitem_line_properties = glade_xml_get_widget(gladexml_popup, "menuitem_line_properties");
   g_signal_connect(plot->menuitem_line_properties, "activate", G_CALLBACK(prefs_line_properties_clicked), plot);
   g_signal_connect(plot->treeview_prefs, "button-press-event", G_CALLBACK(prefs_clicked), plot);
   g_signal_connect(plot->treeview_prefs, "popup-menu", G_CALLBACK(prefs_popup), plot);
}
