/*
 * Plot - plot time-based data on screen and to file with interactive controls
 * Copyright (C) 2006  Jonas Berlin <xkr47@outerspace.dyndns.org>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 ****
 * eventhandlers - gtk event handlers, mostly calls other methods to
 * do the job
 */

#include <gtk/gtkbutton.h>
#include <gtk/gtkcheckbutton.h>
#include <gtk/gtkcheckmenuitem.h>
#include <gtk/gtkdnd.h>
#include <gtk/gtkmenuitem.h>
#include <gtk/gtkrange.h>
#include <gtk/gtkselection.h>
#include <gtk/gtktogglebutton.h>

#include "../plot_draw.xh"
#include "../system.xh"
#include "eventhandlers-plot-draw.xh"
#include "eventhandlers-plot-keyboard.xh"
#include "eventhandlers-plot-mouse.xh"
#include "eventhandlers-prefs.xh"
#include "eventhandlers.xh"
#include "export.xh"
#include "file_open_dialog.xh"
#include "font_dialog.xh"
#include "gtkutil.xh"
#include "plot2-datafiles.xh"
#include "plot2-markers.xh"
#include "plot2-zoom-history.xh"
#include "plot2.xh"
#include "remove_data_file_dialog.xh"
#include "sample_density_dialog.xh"
#include "tail_fam.xh"

#if 0
#define pprintf printf
#else
#define pprintf(...)
#endif

static void about_clicked(GtkMenuItem *about_menuitem, my_plot_t *plot) {
   glade_xml_new(plot->gladefile, "aboutdialog1", NULL);
}

static void add_data_file_clicked(GtkMenuItem *menuitem, my_plot_t *plot) {
   file_open_dialog_run(plot);
}

static void remove_data_file_clicked(GtkMenuItem *menuitem, my_plot_t *plot) {
   remove_data_file_dialog_run(plot);
}

static void remove_all_data_files_clicked(GtkMenuItem *menuitem, my_plot_t *plot) {
   remove_all_data_files(plot);
}

static void refresh_clicked(GtkMenuItem *menuitem, my_plot_t *plot) {
   refresh_all_files(plot);
}

#ifdef USE_FAM

static void menuitem_file_autorefresh_clicked(GtkCheckMenuItem *menuitem, my_plot_t *plot) {
   gboolean newstate = gtk_check_menu_item_get_active(menuitem);
   if(newstate) {
      enable_fam(plot);
   } else {
      disable_fam(plot);
   }
}

static void checkbutton_tail_clicked(GtkCheckButton *checkbutton, my_plot_t *plot) {
   gboolean newstate = gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(checkbutton));
   if(newstate) {
      enable_fam(plot);
   } else {
      disable_fam(plot);
   }
}

#endif

static void button_zoom_prev_clicked(GtkWidget *button_or_menuitem, my_plot_t *plot) {
   switch_to_zoom_history(plot, plot->zoom_pos - 1);
}

static void button_zoom_next_clicked(GtkWidget *button_or_menuitem, my_plot_t *plot) {
   switch_to_zoom_history(plot, plot->zoom_pos + 1);
}

static void button_show_all_clicked(GtkWidget *button_or_menuitem, my_plot_t *plot) {
   // TODO per-view show-all button ?
   for(int vi=0; vi<plot->view_count; ++vi) {
      show_all_nohistory(plot->view[vi]);
   }
   add_to_zoom_history(plot);
}

static void button_zoom_hover(GtkButton *button, my_plot_t *plot) {
   show_zoom_history(plot, plot->zoom_pos + (GTK_WIDGET(button) == plot->button_zoom_prev ? -1 : 1));
}

static void button_zoom_unhover(GtkButton *button, my_plot_t *plot) {
   hide_zoom_history(plot);
}

static void invert_colors_clicked(GtkCheckMenuItem *menuitem, my_plot_t *plot) {
   gboolean newstate = gtk_check_menu_item_get_active(menuitem);

   if(plot->view[0]->draw.inverse_colors == newstate) return;

   for(int vi=0; vi<plot->view_count; ++vi) {
      plot->view[vi]->draw.inverse_colors = newstate;
   }

   reset_gc(plot);

   plot_need_redraw(plot);
}

static void draw_minmax_clicked(GtkCheckMenuItem *menuitem, my_plot_t *plot) {
   boolean_t newstate = gtk_check_menu_item_get_active(menuitem);

   if(plot->view[0]->draw.draw_min_max_lines == newstate) return;

   for(int vi=0; vi<plot->view_count; ++vi) {
      plot->view[vi]->draw.draw_min_max_lines = newstate;
   }

   reset_gc(plot);

   plot_need_redraw(plot);
}

static void draw_average_clicked(GtkCheckMenuItem *menuitem, my_plot_t *plot) {
   boolean_t newstate = gtk_check_menu_item_get_active(menuitem);

   if(plot->view[0]->draw.draw_average_line == newstate) return;

   for(int vi=0; vi<plot->view_count; ++vi) {
      plot->view[vi]->draw.draw_average_line = newstate;
   }

   reset_gc(plot);

   plot_need_redraw(plot);
}

static void button_new_view_clicked(GtkWidget *button_or_menuitem, my_plot_t *plot) {
   plot_add_view(plot);
}

static void button_show_prefs_clicked(GtkWidget *button_or_menuitem, my_plot_t *plot) {
   x_gtk_widget_set_visible(plot->scrolledwindow_prefs, !x_gtk_widget_get_visible(plot->scrolledwindow_prefs));
}

static void font_clicked(GtkMenuItem *menuitem, my_plot_t *plot) {
   font_dialog_run(plot);
}

static void sample_density_clicked(GtkMenuItem *menuitem, my_plot_t *plot) {
   sample_density_dialog_run(plot);
}

static void show_scrollbars_and_tools_clicked(GtkCheckMenuItem *menuitem, my_plot_t *plot) {
   gboolean show = gtk_check_menu_item_get_active(menuitem);
   x_gtk_widget_set_visible(plot->box_xscrolls, show);
   for(int vi=0; vi<plot->view_count; ++vi) {
      x_gtk_widget_set_visible(plot->view[vi]->box_yscrolls, show);
   }
   x_gtk_widget_set_visible(plot->table_tools, show);
   x_gtk_widget_set_visible(plot->scrolledwindow_prefs, show);
}

static void show_disabled_in_legend_clicked(GtkCheckMenuItem *menuitem, my_plot_t *plot) {
   gboolean show = gtk_check_menu_item_get_active(menuitem);
   for(int vi=0; vi<plot->view_count; ++vi) {
      plot->view[vi]->draw.show_disabled = show;
   }
   plot_need_redraw(plot);
}

static void show_legend_left_clicked(GtkCheckMenuItem *menuitem, my_plot_t *plot) {
   for(int vi=0; vi<plot->view_count; ++vi) {
      plot->view[vi]->draw.legend.show = LEGEND_SHOW_LEFT;
   }
   plot_need_redraw(plot);
}

static void show_legend_right_clicked(GtkCheckMenuItem *menuitem, my_plot_t *plot) {
   for(int vi=0; vi<plot->view_count; ++vi) {
      plot->view[vi]->draw.legend.show = LEGEND_SHOW_RIGHT;
   }
   plot_need_redraw(plot);
}

static void hide_legend_clicked(GtkCheckMenuItem *menuitem, my_plot_t *plot) {
   for(int vi=0; vi<plot->view_count; ++vi) {
      plot->view[vi]->draw.legend.show = LEGEND_HIDE;
   }
   plot_need_redraw(plot);
}

static void show_line_markers_clicked(GtkCheckMenuItem *menuitem, my_plot_t *plot) {
   gboolean show = gtk_check_menu_item_get_active(menuitem);
   plot->show_line_markers = show;
   show_markers(plot, plot->view[0]->vmarkers_drawn);
}

static void fullscreen_clicked(GtkCheckMenuItem *menuitem, my_plot_t *plot) {
   if(plot->is_fullscreen) {
      gtk_window_unfullscreen(plot->parent);
   } else {
      gtk_window_fullscreen(plot->parent);
   }
   plot->is_fullscreen = !plot->is_fullscreen;
}

static void doublefont_clicked(GtkCheckMenuItem *menuitem, my_plot_t *plot) {
   plot->double_font_size = !plot->double_font_size;
   plot_size_or_font_changed(plot);
}

static void plot_generic_scrollbar_value_changed(my_plot_t *plot) {
   double xlength = data_scrollbar_controller_get_page_length(plot->dsc_x);
   double ylength = data_scrollbar_controller_get_page_length(plot->view[0]->dsc_y);

   double xlo, xhi;
   double ylo, yhi;
   data_scrollbar_controller_get_range(plot->dsc_x, &xlo, &xhi);
   data_scrollbar_controller_get_range(plot->view[0]->dsc_y, &ylo, &yhi); // TODO multiview

   for(int vi=0; vi<plot->view_count; ++vi) {
      plot->view[vi]->draw.xstart = (my_time_t)xlo;
      plot->view[vi]->draw.xend = (my_time_t)xhi;
   }

   gboolean show_all_enable =
      (xhi-xlo - xlength) / (xhi - xlo) > 1e-10 ||
      (yhi-ylo - ylength) / (yhi - ylo) > 1e-10;

   if(plot->button_show_all) {
      gtk_widget_set_sensitive(plot->button_show_all, show_all_enable);
   }
   if(plot->menuitem_show_all) {
      gtk_widget_set_sensitive(plot->menuitem_show_all, show_all_enable);
   }

   delayed_add_to_zoom_history(plot);
}

static void plot_xscrollbar_value_changed(GtkRange *not_always_provided, my_plot_t *plot) {

   double xlength = data_scrollbar_controller_get_page_length(plot->dsc_x);
   double xoff = data_scrollbar_controller_get_page_off(plot->dsc_x);

   for(int vi=0; vi<plot->view_count; ++vi) {
      plot->view[vi]->draw.xmin = xoff;
      plot->view[vi]->draw.xmax = xoff + xlength;
   }

   plot_generic_scrollbar_value_changed(plot);

   plot_need_redraw(plot);
}

static void plot_yscrollbar_value_changed(GtkRange *not_always_provided, my_plot_view_t *view) {

   double ylength = data_scrollbar_controller_get_page_length(view->dsc_y);
   double yoff = data_scrollbar_controller_get_page_off(view->dsc_y);

   view->draw.ymin = yoff;
   view->draw.ymax = yoff + ylength;

   plot_generic_scrollbar_value_changed(view->parent);

   plot_view_need_redraw(view);
}

typedef enum { DND_INFO_URI, DND_INFO_TEXT } my_drag_drop_info_t;

static void drag_n_drop(GtkWidget *widget, GdkDragContext *context, gint x, gint y, GtkSelectionData *data, guint info, guint drag_time, my_plot_t *plot) {
   switch(info) {
      case DND_INFO_URI: {
	 gchar **uris = gtk_selection_data_get_uris(data);
	 if(!uris) break;
	 for(int i=0; uris[i]; ++i) {
	    printf("Got %s\n", uris[i]);
	    char *hostname;
	    char *path = g_filename_from_uri(uris[i], &hostname, NULL);
	    if(path && (!hostname || !strncmp(hostname, "localhost", (size_t)9))) {
	       add_datafile(plot, path);
	    }
	    if(path) free(path);
	    if(hostname) free(hostname);
	 }
	 g_strfreev (uris);
	 break;
      }
      case DND_INFO_TEXT: {
	 gchar *text = (gchar *)gtk_selection_data_get_text(data);
	 if(!text) break;
	 // assume text is a set of lines, each line containing a filename
	 gchar **paths = g_strsplit_set(text, "\r\n", 0);
	 for(int i=0; paths[i]; ++i) {
	    if(strspn(paths[i], " \t") == strlen(paths[i])) continue; // only whitespace, if any
	    add_datafile(plot, paths[i]);
	 }
	 g_strfreev(paths);
	 free(text);
	 break;
      }
   }
}


void plot_view_connect_signals(my_plot_view_t *view, GladeXML *gladexml) {

   // setup drag and drop

   gtk_drag_dest_set(view->da, GTK_DEST_DEFAULT_ALL, NULL, 0, GDK_ACTION_COPY);
   GtkTargetList *target_list = gtk_target_list_new(NULL, 0);
   gtk_target_list_add_uri_targets(target_list, DND_INFO_URI);
   gtk_target_list_add_text_targets(target_list, DND_INFO_TEXT);
   gtk_drag_dest_set_target_list(view->da, target_list);
   gtk_target_list_unref(target_list);
   g_signal_connect(view->da, "drag-data-received", G_CALLBACK(drag_n_drop), view);

   // drawing area

   gtk_widget_add_events(view->da, GDK_KEY_PRESS_MASK | GDK_KEY_RELEASE_MASK);
   g_signal_connect(view->da, "key-press-event", G_CALLBACK(plot_key_event), view);
   g_signal_connect(view->da, "key-release-event", G_CALLBACK(plot_key_event), view);
   g_signal_connect(view->da, "focus-in-event", G_CALLBACK(plot_focus_event_before), view);
   g_signal_connect_after(view->da, "focus-in-event", G_CALLBACK(plot_focus_event_after), view);
   g_signal_connect(view->da, "focus-out-event", G_CALLBACK(plot_focus_event_before), view);
   g_signal_connect_after(view->da, "focus-out-event", G_CALLBACK(plot_focus_event_after), view);

   gtk_widget_add_events(view->da, GDK_ENTER_NOTIFY_MASK | GDK_LEAVE_NOTIFY_MASK);
   g_signal_connect(view->da, "enter-notify-event", G_CALLBACK(plot_enter_notify_event), view);
   g_signal_connect(view->da, "leave-notify-event", G_CALLBACK(plot_leave_notify_event), view);

   gtk_widget_add_events(view->da, GDK_POINTER_MOTION_MASK);
   g_signal_connect(view->da, "motion-notify-event", G_CALLBACK(plot_motion_notify_event), view);

   gtk_widget_add_events(view->da, GDK_BUTTON_PRESS_MASK);
   g_signal_connect(view->da, "button-press-event", G_CALLBACK(plot_button_press_event), view);

   gtk_widget_add_events(view->da, GDK_BUTTON_RELEASE_MASK);
   g_signal_connect(view->da, "button-release-event", G_CALLBACK(plot_button_release_event), view);


   g_signal_connect(view->da, "map-event", G_CALLBACK(plot_map_event), view);
   g_signal_connect(view->da, "configure-event", G_CALLBACK(plot_configure_event), view);
   g_signal_connect(view->da, "expose-event", G_CALLBACK(plot_expose_event), view);

   // y scrollbars

   g_signal_connect(view->dsc_y->scroll_off, "value-changed", G_CALLBACK(plot_yscrollbar_value_changed), view);
   g_signal_connect(view->dsc_y->scroll_off, "changed", G_CALLBACK(plot_yscrollbar_value_changed), view);

   view->box_yscrolls = glade_xml_get_widget(gladexml, "box_yscrolls");

   plot_yscrollbar_value_changed(NULL, view);
}

void plot_connect_signals(my_plot_t *plot, GladeXML *gladexml) {

   GdkScreen* screen = gtk_window_get_screen(plot->parent);
   g_signal_connect(screen, "size-changed", G_CALLBACK(screen_size_changed_event), plot);

/*
  GdkRectangle r;
  gdk_screen_get_monitor_geometry(screen, 0, &r);

  printf("screen %d,%d\n", gdk_screen_get_width(screen), gdk_screen_get_height(screen));
  printf("  mm   %d,%d\n", gdk_screen_get_width_mm(screen), gdk_screen_get_height_mm(screen));
  printf("  moni %d,%d\n", r.width, r.height);
*/

   // buttons

   plot->button_show_all = glade_xml_get_widget(gladexml, "button_show_all");
   g_signal_connect(plot->button_show_all, "clicked", G_CALLBACK(button_show_all_clicked), plot);

   g_signal_connect(plot->button_zoom_prev, "clicked", G_CALLBACK(button_zoom_prev_clicked), plot);
   g_signal_connect(plot->button_zoom_next, "clicked", G_CALLBACK(button_zoom_next_clicked), plot);
   g_signal_connect(plot->button_zoom_prev, "enter", G_CALLBACK(button_zoom_hover), plot);
   g_signal_connect(plot->button_zoom_next, "enter", G_CALLBACK(button_zoom_hover), plot);
   g_signal_connect(plot->button_zoom_prev, "leave", G_CALLBACK(button_zoom_unhover), plot);
   g_signal_connect(plot->button_zoom_next, "leave", G_CALLBACK(button_zoom_unhover), plot);

   plot->box_plots = glade_xml_get_widget(gladexml, "vbox_plot");
   //g_signal_connect(glade_xml_get_widget(gladexml, "button_new_view"), "clicked", G_CALLBACK(button_new_view_clicked), plot);

   g_signal_connect(glade_xml_get_widget(gladexml, "button_show_prefs"), "clicked", G_CALLBACK(button_show_prefs_clicked), plot);

   //g_signal_connect(glade_xml_get_widget(gladexml, "button_menu"), "clicked", G_CALLBACK(menu_clicked), glade_xml_get_widget(gladexml, "menubar_main"));

   // fps label

   plot->label_fps = GTK_LABEL(glade_xml_get_widget(gladexml, "label_fps"));

   // menu

   plot->box_xscrolls = glade_xml_get_widget(gladexml, "box_xscrolls");
   plot->table_tools = glade_xml_get_widget(gladexml, "table_tools");
   plot->treeview_prefs = glade_xml_get_widget(gladexml, "treeview_prefs");
   plot->scrolledwindow_prefs = glade_xml_get_widget(gladexml, "scrolledwindow_prefs");

   prefs_init(plot);

#ifdef USE_FAM

   plot->menuitem_file_refresh = glade_xml_get_widget(gladexml, "menuitem_file_refresh");
   plot->menuitem_file_autorefresh = glade_xml_get_widget(gladexml, "menuitem_file_autorefresh");
   plot->checkbutton_tail = glade_xml_get_widget(gladexml, "checkbutton_tail");
   plot->checkbutton_tail_move = glade_xml_get_widget(gladexml, "checkbutton_tail_move");

   plot->xscroll_off = glade_xml_get_widget(gladexml, "xscroll_scrollbar_off");
   plot->xscroll_off_rel = glade_xml_get_widget(gladexml, "xscroll_scrollbar_off_rel");
   plot->xscroll_zoompoint = glade_xml_get_widget(gladexml, "xscroll_scale");

   gtk_widget_set_sensitive(plot->menuitem_file_autorefresh, 1);
   gtk_widget_set_sensitive(plot->checkbutton_tail, 1);
   gtk_widget_set_sensitive(plot->checkbutton_tail_move, 1);

   g_signal_connect(plot->menuitem_file_autorefresh, "activate", G_CALLBACK(menuitem_file_autorefresh_clicked), plot);
   g_signal_connect(plot->checkbutton_tail, "toggled", G_CALLBACK(checkbutton_tail_clicked), plot);

#endif

   g_signal_connect(glade_xml_get_widget(gladexml, "menuitem_file_add_data_file"), "activate", G_CALLBACK(add_data_file_clicked), plot);
   g_signal_connect(glade_xml_get_widget(gladexml, "menuitem_file_remove_data_file"), "activate", G_CALLBACK(remove_data_file_clicked), plot);
   g_signal_connect(glade_xml_get_widget(gladexml, "menuitem_file_remove_all_data_files"), "activate", G_CALLBACK(remove_all_data_files_clicked), plot);
   g_signal_connect(glade_xml_get_widget(gladexml, "menuitem_file_refresh"), "activate", G_CALLBACK(refresh_clicked), plot);
   g_signal_connect(glade_xml_get_widget(gladexml, "menuitem_file_export_as"), "activate", G_CALLBACK(export_as_clicked), plot);
   g_signal_connect(glade_xml_get_widget(gladexml, "menuitem_quit"), "activate", G_CALLBACK(request_quit), (gpointer)2);

   g_signal_connect(glade_xml_get_widget(gladexml, "menuitem_view_previous"), "activate", G_CALLBACK(button_zoom_prev_clicked), plot);
   g_signal_connect(glade_xml_get_widget(gladexml, "menuitem_view_next"), "activate", G_CALLBACK(button_zoom_next_clicked), plot);

   plot->menuitem_show_all = glade_xml_get_widget(gladexml, "menuitem_show_all");
   g_signal_connect(plot->menuitem_show_all, "activate", G_CALLBACK(button_show_all_clicked), plot);
   g_signal_connect(glade_xml_get_widget(gladexml, "menuitem_new_view"), "activate", G_CALLBACK(button_new_view_clicked), plot);
   g_signal_connect(glade_xml_get_widget(gladexml, "menuitem_fullscreen"), "activate", G_CALLBACK(fullscreen_clicked), plot);
   g_signal_connect(glade_xml_get_widget(gladexml, "menuitem_doublefont"), "activate", G_CALLBACK(doublefont_clicked), plot);

   g_signal_connect(glade_xml_get_widget(gladexml, "menuitem_inverse_colors"), "activate", G_CALLBACK(invert_colors_clicked), plot);
   g_signal_connect(glade_xml_get_widget(gladexml, "menuitem_font"), "activate", G_CALLBACK(font_clicked), plot);
   g_signal_connect(glade_xml_get_widget(gladexml, "menuitem_sample_density"), "activate", G_CALLBACK(sample_density_clicked), plot);
   g_signal_connect(glade_xml_get_widget(gladexml, "menuitem_show_legend_left"), "activate", G_CALLBACK(show_legend_left_clicked), plot);
   g_signal_connect(glade_xml_get_widget(gladexml, "menuitem_show_legend_right"), "activate", G_CALLBACK(show_legend_right_clicked), plot);
   g_signal_connect(glade_xml_get_widget(gladexml, "menuitem_hide_legend"), "activate", G_CALLBACK(hide_legend_clicked), plot);
   g_signal_connect(glade_xml_get_widget(gladexml, "menuitem_draw_minmax"), "activate", G_CALLBACK(draw_minmax_clicked), plot);
   g_signal_connect(glade_xml_get_widget(gladexml, "menuitem_draw_average"), "activate", G_CALLBACK(draw_average_clicked), plot);
   g_signal_connect(glade_xml_get_widget(gladexml, "menuitem_show_line_markers"), "activate", G_CALLBACK(show_line_markers_clicked), plot);
   g_signal_connect(glade_xml_get_widget(gladexml, "menuitem_show_scrollbars_and_tools"), "activate", G_CALLBACK(show_scrollbars_and_tools_clicked), plot);
   g_signal_connect(glade_xml_get_widget(gladexml, "menuitem_show_disabled_in_legend"), "activate", G_CALLBACK(show_disabled_in_legend_clicked), plot);

   g_signal_connect(glade_xml_get_widget(gladexml, "menuitem_about"), "activate", G_CALLBACK(about_clicked), plot);

   // x scrollbars

   g_signal_connect(plot->dsc_x->scroll_off, "value-changed", G_CALLBACK(plot_xscrollbar_value_changed), plot);
   g_signal_connect(plot->dsc_x->scroll_off, "changed", G_CALLBACK(plot_xscrollbar_value_changed), plot);

   ++plot->disable_delayed_add_zoom;

   plot_xscrollbar_value_changed(NULL, plot);

   --plot->disable_delayed_add_zoom;
}

/*
 * Local variables:
 * c-file-style: "ellemtel"
 * c-file-offsets: ((c . c-lineup-dont-change) (statement-cont . (lambda (le) (if (save-excursion (goto-char (cdr le)) (looking-at "return")) (c-lineup-java-inher le) (c-lineup-math le)))))
 * End:
 */
