/*
 * Plot - plot time-based data on screen and to file with interactive controls
 * Copyright (C) 2006  Jonas Berlin <xkr47@outerspace.dyndns.org>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 ****
 * file_open_dialog - functions related to "file open" dialog
 */

#include <gtk/gtkdrawingarea.h>
#include <gtk/gtkfilechooserdialog.h>
#include <gtk/gtktable.h>

#include <math.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <unistd.h>

#include "../system.xh"
#include "file_open_dialog.xh"
#include "paintapi-gtk.xh"
#include "plot2-datafiles.xh"

static const char *labelNames[] = {
   N_("File date:"),
   N_("Start time:"),
   N_("End time:"),
   N_("Covered time:"),
   N_("Time format:"),
   N_("Data columns:"),
};

typedef struct {
      GtkWidget *da;
      GtkLabel *labels[SZ(labelNames)];
      paintapi_t *paintapi;
      my_plot_draw_t draw;
} adf_data_t;


static void adf_update(GtkFileChooser *chooser, adf_data_t *data) {
   char *filename = gtk_file_chooser_get_preview_filename(chooser);
   my_dataset_t *dataset = scan_dataset(filename, 10);
   gtk_file_chooser_set_preview_widget_active(chooser, 1);

   plot_draw_remove_all_data_files(&data->draw);

   if(!dataset) {
      struct stat st;
      if(!stat(filename, &st)) {
	 gtk_label_set_text(data->labels[0], format_time(st.st_mtime * TIME_MULTIPLIER, 0));
      } else {
	 gtk_label_set_text(data->labels[0], "xxxx-xx-xx xx:xx:xx");
      }
      gtk_label_set_text(data->labels[1], "");
      gtk_label_set_text(data->labels[2], "");
      gtk_label_set_text(data->labels[3], "");
      gtk_label_set_text(data->labels[4], "");
      gtk_label_set_text(data->labels[5], "");
      gtk_widget_queue_draw(data->da);

      g_free(filename);
      return;
   }

   char buf[40];
   int idx = 0;

   gtk_label_set_text(data->labels[idx++], format_time(dataset->last_modified, 0));
   gtk_label_set_text(data->labels[idx++], format_time(dataset->ts, 0));
   gtk_label_set_text(data->labels[idx++], format_time(dataset->te, 0));
   gtk_label_set_text(data->labels[idx++], format_difftime(dataset->te - dataset->ts, 0));
   gtk_label_set_text(data->labels[idx++], dataset->timetype == TIMETYPE_EPOCH ? _("epoch") : _("relative"));
   snprintf(buf, sizeof(buf), "%d", dataset->item_count);
   gtk_label_set_text(data->labels[idx++], buf);

   unref_dataset(dataset);

   // read preview

   dataset = read_dataset_preview(filename, 400);

   g_free(filename);

   if(!dataset) {
      return;
   }

   my_plot_line_info_set_arr_set_length(data->draw.line_set, 1);

   data->draw.line_set[0].dataset = dataset;

   data->draw.xmin = data->draw.xstart = dataset->ts;
   data->draw.xmax = data->draw.xend = dataset->te;

   double vmin = INFINITY, vmax = -INFINITY;

   for(int dsi=0; dsi<dataset->item_count; ++dsi) {
      int li = data->draw.line_set[0].line_count;
      my_plot_line_info_arr_set_length(data->draw.line_set[0].line, li+1);
      int linecoloridx = dsi;
      data->draw.line_set[0].line[li].color = default_line_colors[linecoloridx % num_default_line_colors];
      data->draw.line_set[0].line[li].dataset_idx = dsi;
      data->draw.line_set[0].line[li].enabled = 1;
      data->draw.line_set[0].line[li].mode = LINE_MODE_NORMAL;
      if(dsi == 0 || vmin > dataset->item[dsi].min) vmin = dataset->item[dsi].min;
      if(dsi == 0 || vmax < dataset->item[dsi].max) vmax = dataset->item[dsi].max;
   }

   data->draw.ymin = vmin - (vmax-vmin)/10;
   data->draw.ymax = vmax + (vmax-vmin)/10;

   gtk_widget_queue_draw(data->da);
}

static gboolean adtd_map_event(GtkWidget *widget, GdkEvent *do_not_use, adf_data_t *data) {
   data->draw.paintapi = paintapi_gtk_new(data->da->window, &data->draw);

   // set the background part 1 - create new style for widget
   GtkStyle *style = gtk_widget_get_style(data->da);
   style = gtk_style_copy(style);
   gtk_widget_set_style(data->da, style);

   // set the background part 2
   GdkColor bg;
   bg.red = bg.green = bg.blue = 0;
   gdk_colormap_alloc_color(gdk_colormap_get_system(), &bg, 0, 0);
   gdk_window_set_background(data->da->window, &bg);

   // fonts
   data->draw.legend.font = zmalloc_c(sizeof(paintapi_font_t));
   data->draw.grid.font_xaxis = data->draw.legend.font;
   data->draw.grid.font_yaxis = data->draw.legend.font;
   data->draw.legend.font->ctx = gtk_widget_get_pango_context(data->da);
   data->draw.legend.font->font = pango_font_description_from_string("Arial 9");
   return 1;
}

static gboolean adtd_configure_event(GtkWidget *widget, GdkEventConfigure *do_not_use, adf_data_t *data) {
   int w,h;
   gdk_drawable_get_size(data->da->window, &w, &h);
   data->draw.plot_xe = w - 1;
   data->draw.plot_ye = h - 1;
   return 1;
}

static gboolean adtd_expose_event(GtkWidget *widget, GdkEventExpose *event, adf_data_t *data) {
   if(data->draw.paintapi && data->draw.line_set_count)
      plot_draw(&data->draw);
   return 1;
}

void file_open_dialog_run(my_plot_t *plot) {
   GladeXML *gladexml = glade_xml_new(plot->gladefile, "filechooserdialog_add_file", NULL);
   GtkWidget *dialog = glade_xml_get_widget(gladexml, "filechooserdialog_add_file");

   GtkFileChooser *chooser = GTK_FILE_CHOOSER(dialog);

   gtk_file_chooser_set_select_multiple(chooser, 1);

   GtkWidget *table = gtk_table_new(5, 2, 0);
   gtk_table_set_col_spacings(GTK_TABLE(table), 2);

#define setcolor(x,r_,g_,b_) x.r=r_;x.g=g_;x.b=b_;

   adf_data_t adtd;
   bzero(&adtd, sizeof(adtd));

   // plot_draw init
   plot_draw_init(&adtd.draw);

   setcolor(adtd.draw.grid.minor_color, 0x6060, 0x6060, 0x6060);
   setcolor(adtd.draw.grid.major_color, 0xc0c0, 0xc0c0, 0xc0c0);
   setcolor(adtd.draw.grid.zerogrid_color, 0xffff, 0x6060, 0x4040);
   setcolor(adtd.draw.grid.sep_color, 0x8080, 0x8080, 0x8080);
   setcolor(adtd.draw.grid.label_color, 0xffff, 0xffff, 0xffff);
   setcolor(adtd.draw.legend.bg_color, 0x2020, 0x2020, 0x2020);
   setcolor(adtd.draw.legend.border_color, 0xa0a0, 0xa0a0, 0xa0a0);
   setcolor(adtd.draw.legend.text_color, 0xe0e0, 0xe0e0, 0xe0e0);

   adtd.draw.legend.show = LEGEND_HIDE;

   // /plot_draw init

   adtd.da = gtk_drawing_area_new();
   gtk_widget_set_size_request(adtd.da, 80, 80);
   gtk_table_attach(GTK_TABLE(table), adtd.da, 0, 2, 0, 1, GTK_EXPAND|GTK_FILL, GTK_EXPAND|GTK_FILL, 0, 0);
   g_signal_connect(adtd.da, "map-event", G_CALLBACK(adtd_map_event), &adtd);
   g_signal_connect(adtd.da, "configure-event", G_CALLBACK(adtd_configure_event), &adtd);
   g_signal_connect(adtd.da, "expose-event", G_CALLBACK(adtd_expose_event), &adtd);

   for(unsigned i=0; i<SZ(labelNames); ++i) {
      GtkWidget *labelNameW = gtk_label_new(_(labelNames[i]));
      GtkWidget *labelW = gtk_label_new("?");

      gtk_misc_set_alignment(GTK_MISC(labelNameW), 1.0, 0.5);
      gtk_misc_set_alignment(GTK_MISC(labelW), 0.0, 0.5);

      adtd.labels[i] = GTK_LABEL(labelW);

      gtk_table_attach(GTK_TABLE(table), labelNameW, 0, 1, i+1, i+2, GTK_EXPAND|GTK_FILL, 0, 0, 0);
      gtk_table_attach(GTK_TABLE(table), labelW,     1, 2, i+1, i+2, GTK_EXPAND|GTK_FILL, 0, 0, 0);
   }
   gtk_widget_show_all(table);

   gtk_file_chooser_set_preview_widget(chooser, table);
   g_signal_connect(chooser, "update-preview", G_CALLBACK(adf_update), &adtd);

   GtkFileFilter *filter;

   filter = gtk_file_filter_new();
   gtk_file_filter_set_name(filter, _("Text files"));
   gtk_file_filter_add_mime_type(filter, "text/plain");
   gtk_file_chooser_add_filter(chooser, filter);

   filter = gtk_file_filter_new();
   gtk_file_filter_set_name(filter, _("All files"));
   gtk_file_filter_add_pattern(filter, "*");
   gtk_file_chooser_add_filter(chooser, filter);

   if(gtk_dialog_run(GTK_DIALOG(dialog)) == GTK_RESPONSE_OK) {
      gtk_widget_hide(dialog);

      GSList* files;
      for(files = gtk_file_chooser_get_filenames(chooser); files; files = files->next) {
	 add_datafile(plot, files->data);
      }
      g_slist_free(files);
   }

   plot_draw_deinit(&adtd.draw);

   gtk_widget_destroy(dialog);
}

/*
 * Local variables:
 * c-file-style: "ellemtel"
 * c-file-offsets: ((c . c-lineup-dont-change) (statement-cont . (lambda (le) (if (save-excursion (goto-char (cdr le)) (looking-at "return")) (c-lineup-java-inher le) (c-lineup-math le)))))
 * End:
 */
