/*
 * Plot - plot time-based data on screen and to file with interactive controls
 * Copyright (C) 2006  Jonas Berlin <xkr47@outerspace.dyndns.org>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 ****
 * plot2-markers - functions related to marker management (markers are
 * dotted lines on plot indicating cursor and other positions)
 */

#include <math.h>

#include "plot2-markers.xh"

const char cursor_marker_dashes[2] = { 2, 4 };

void set_and_show_markers(my_plot_view_t *view, int x, int y) {
   if(!view->draw.grid.minor_gc) {
      // gui not yet initialized
      return;
   }

   destroy_markers(view->parent);

   if(y > view->draw.plot_yde) return;

   view->paintapi_direct->gc_set_dashes(view->paintapi_direct, view->parent->cursor_marker_gc, -x, cursor_marker_dashes, SZ(cursor_marker_dashes));
   view->paintapi_direct->draw_line(view->paintapi_direct, view->parent->cursor_marker_gc, 0, y, view->draw.plot_xe, y);
   view->hmarkers_drawn = 1;
   view->hmarkers_set = 1;

   int yo = 0;
   for(int vi=0; vi<view->parent->view_count; ++vi) {
      my_plot_view_t *v2 = view->parent->view[vi];
      v2->paintapi_direct->gc_set_dashes(v2->paintapi_direct, v2->parent->cursor_marker_gc, -y, cursor_marker_dashes, SZ(cursor_marker_dashes));
      v2->paintapi_direct->draw_line(v2->paintapi_direct, v2->parent->cursor_marker_gc, x, 0, x, v2->draw.plot_yde);
      v2->hline = y;
      v2->vline = x;
      v2->vmarkers_drawn = 1;
      v2->vmarkers_set = 1;
      yo += v2->draw.plot_ye + 1;
   }

   double xlength = data_scrollbar_controller_get_page_length(view->parent->dsc_x);
   double xoff = data_scrollbar_controller_get_page_off(view->parent->dsc_x);

   double ylength = data_scrollbar_controller_get_page_length(view->dsc_y);
   double yoff = data_scrollbar_controller_get_page_off(view->dsc_y);

   my_time_t xv = (my_time_t)(x * xlength / view->draw.plot_xe + xoff);
   for(int lsi=0; lsi<view->draw.line_set_count; ++lsi) {
      int p = my_time_arr_find(view->draw.line_set[lsi].dataset->time, xv, 0);
      for(int li=0; li<view->draw.line_set[lsi].line_count; ++li) {
	 if(!view->draw.line_set[lsi].line[li].enabled) continue;

	 int dataset_idx = view->draw.line_set[lsi].line[li].dataset_idx;

	 int dy = !__isnan(view->draw.line_set[lsi].dataset->item[dataset_idx].data[p]) ?
		  (int)(view->draw.plot_yde - (view->draw.line_set[lsi].dataset->item[dataset_idx].data[p] - yoff) / ylength * view->draw.plot_yde) :
		  -1;

	 if(dy >= 0) {
	    view->draw.line_set[lsi].line[li].hline = dy;
	    if(view->parent->show_line_markers) {
	       view->paintapi_direct->draw_line(view->paintapi_direct, view->draw.line_set[lsi].line[li].horiz_marker_gc, 0, dy, view->draw.plot_xe, dy);
	       view->draw.line_set[lsi].line[li].markers_drawn = 1;
	    }
	 }
      }
   }
}

void show_view_markers(my_plot_view_t *view, boolean_t show, boolean_t direct) {
   paintapi_t *api = direct ? view->paintapi_direct : view->draw.paintapi;
   if(view->hmarkers_drawn != show && view->hmarkers_set) {
      view->hmarkers_drawn = show;
      api->gc_set_dashes(api, view->parent->cursor_marker_gc, -view->vline, cursor_marker_dashes, SZ(cursor_marker_dashes));
      api->draw_line(api, view->parent->cursor_marker_gc, 0, view->hline, view->draw.plot_xe, view->hline);
   }
   if(view->vmarkers_drawn != show && view->vmarkers_set) {
      view->vmarkers_drawn = show;
      api->gc_set_dashes(api, view->parent->cursor_marker_gc, -view->hline, cursor_marker_dashes, SZ(cursor_marker_dashes));
      api->draw_line(api, view->parent->cursor_marker_gc, view->vline, 0, view->vline, view->draw.plot_yde);
   }

   show = show && view->parent->show_line_markers;

   for(int lsi=0; lsi<view->draw.line_set_count; ++lsi) {
      for(int li=0; li<view->draw.line_set[lsi].line_count; ++li) {
	 if(view->draw.line_set[lsi].line[li].markers_drawn != show && view->draw.line_set[lsi].line[li].hline != -1) {
	    view->draw.line_set[lsi].line[li].markers_drawn = show;
	    api->draw_line(api,
			   view->draw.line_set[lsi].line[li].horiz_marker_gc,
			   0, view->draw.line_set[lsi].line[li].hline,
			   view->draw.plot_xe, view->draw.line_set[lsi].line[li].hline);
	 }
      }
   }
}

void show_markers(my_plot_t *plot, boolean_t show) {
   for(int vi=0; vi<plot->view_count; ++vi) {
      show_view_markers(plot->view[vi], show, true);
   }
}

void restore_markers_after_draw(my_plot_view_t *view) {
   reset_view_markers(view);
   show_view_markers(view, true, false);
}

// mark markers as "not drawn"
void reset_view_markers(my_plot_view_t *view) {
   view->hmarkers_drawn = 0;
   view->vmarkers_drawn = 0;
   for(int lsi=0; lsi<view->draw.line_set_count; ++lsi) {
      for(int li=0; li<view->draw.line_set[lsi].line_count; ++li) {
	 view->draw.line_set[lsi].line[li].markers_drawn = 0;
      }
   }
}

void reset_markers(my_plot_t *plot) {
   for(int vi=0; vi<plot->view_count; ++vi) {
      reset_view_markers(plot->view[vi]);
   }
}

void destroy_view_markers(my_plot_view_t *view) {
   show_view_markers(view, false, true);
   view->hmarkers_set = 0;
   view->vmarkers_set = 0;
   for(int lsi=0; lsi<view->draw.line_set_count; ++lsi) {
      for(int li=0; li<view->draw.line_set[lsi].line_count; ++li) {
	 view->draw.line_set[lsi].line[li].hline = -1;
      }
   }
}

void destroy_markers(my_plot_t *plot) {
   for(int vi=0; vi<plot->view_count; ++vi) {
      destroy_view_markers(plot->view[vi]);
   }
}

/*
 * Local variables:
 * c-file-style: "ellemtel"
 * c-file-offsets: ((c . c-lineup-dont-change) (statement-cont . (lambda (le) (if (save-excursion (goto-char (cdr le)) (looking-at "return")) (c-lineup-java-inher le) (c-lineup-math le)))))
 * End:
 */
