/*
 * Plot - plot time-based data on screen and to file with interactive controls
 * Copyright (C) 2006  Jonas Berlin <xkr47@outerspace.dyndns.org>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 ****
 * plot2-zoom-history - functions for managing the zoom history of a
 * plot
 */

#include <math.h>

#include "plot2-markers.xh"
#include "plot2-zoom-history.xh"
#include "tail_fam.xh"

#if 0
#define pprintf printf
#else
#define pprintf(...)
#endif

static void atzh_helper(my_zoom_view_t *z, data_scrollbar_controller_t *dsc) {
   z->off = data_scrollbar_controller_get_page_off(dsc);
   z->length_log = data_scrollbar_controller_get_page_length_log(dsc);
   z->scalepos = data_scrollbar_controller_get_zoompoint(dsc);
}

void add_to_zoom_history(my_plot_t *plot) {
   if(plot->view[0]->draw.line_set_count == 0) {
      // don't save history if no plots
      return;
   }

   int idx = plot->zoom_pos + 1;
   if(idx > plot->zoom_history_count) idx = plot->zoom_history_count;
   pprintf("add %d\n", idx);

   cancel_delayed_add_to_zoom_history(plot);

   my_zoom_view_arr_set_length(plot->zoom_history, idx + 1);
   atzh_helper(&plot->zoom_history[idx], plot->dsc_x);

   for(int vi=0; vi<plot->view_count; ++vi) {
      atzh_helper(&plot->view[vi]->zoom_history_dflt, plot->view[vi]->dsc_y); // TODO do this in clone instead and get rid of CUSTOM default
      my_zoom_view_arr_set_length(plot->view[vi]->zoom_history, idx + 1);
      atzh_helper(&plot->view[vi]->zoom_history[idx], plot->view[vi]->dsc_y);
   }

   if(idx > 0) {
      gtk_widget_set_sensitive(plot->button_zoom_prev, 1);
   }
   gtk_widget_set_sensitive(plot->button_zoom_next, 0);

   plot->zoom_pos = idx;
}

void switch_to_zoom_history(my_plot_t *plot, int idx) {
   pprintf("swi %d\n", idx);

   if(idx < 0) return;
   if(idx >= plot->zoom_history_count) return;
   if(idx == plot->zoom_pos) return;

   disable_fam(plot);

   ++plot->disable_delayed_add_zoom;

   plot->zoom_pos = idx;

   gtk_widget_set_sensitive(plot->button_zoom_prev, idx > 0);
   gtk_widget_set_sensitive(plot->button_zoom_next, idx < plot->zoom_history_count-1);

   my_zoom_view_t *z = &plot->zoom_history[idx];

   data_scrollbar_controller_set_page_length_log(plot->dsc_x, z->length_log);
   data_scrollbar_controller_set_page_off(plot->dsc_x, z->off);
   data_scrollbar_controller_set_zoompoint(plot->dsc_x, z->scalepos);

   for(int vi=0; vi<plot->view_count; ++vi) {
      z = &plot->view[vi]->zoom_history[idx];

      data_scrollbar_controller_set_page_length_log(plot->view[vi]->dsc_y, z->length_log);
      data_scrollbar_controller_set_page_off(plot->view[vi]->dsc_y, z->off);
      data_scrollbar_controller_set_zoompoint(plot->view[vi]->dsc_y, z->scalepos);

      update_cursor(plot->view[vi]);
   }

   --plot->disable_delayed_add_zoom;
}

void show_zoom_history(my_plot_t *plot, int idx) {
   if(idx < 0) return;
   if(idx >= plot->zoom_history_count) return;

   my_zoom_view_t *z = &plot->zoom_history[idx];

   double xoff = data_scrollbar_controller_get_page_off(plot->dsc_x);
   double xlength = data_scrollbar_controller_get_page_length(plot->dsc_x);

   double z__xoff = z->off;
   double z__xoff2 = z->off + exp(z->length_log);

   for(int vi=0; vi<plot->view_count; ++vi) {
      my_plot_view_t *view = plot->view[vi];

      z = &view->zoom_history[idx];

      double yoff = data_scrollbar_controller_get_page_off(view->dsc_y);
      double ylength = data_scrollbar_controller_get_page_length(view->dsc_y);

      double z__yoff = z->off;
      double z__yoff2 = z->off + exp(z->length_log);

      int x = (int)(view->draw.plot_xe * (z__xoff - xoff) / xlength);
      int y = (int)(view->draw.plot_yde * (z__yoff - yoff) / ylength);
      int x2 = (int)(view->draw.plot_xe * (z__xoff2 - xoff) / xlength);
      int y2 = (int)(view->draw.plot_yde * (z__yoff2 - yoff) / ylength);

      y = view->draw.plot_yde - y;
      y2 = view->draw.plot_yde - y2;

      view->bx = x;
      view->by = y;
      view->box = x2;
      view->boy = y2;

      view->paintapi_direct->gc_set_dashes(view->paintapi_direct, view->parent->cursor_marker_gc, -view->hline, cursor_marker_dashes, SZ(cursor_marker_dashes));
      view->paintapi_direct->draw_rectangle(view->paintapi_direct, view->parent->cursor_marker_gc, 0, x, y, x2, y2);
   }
}

void hide_zoom_history(my_plot_t *plot) {
   for(int vi=0; vi<plot->view_count; ++vi) {
      if(plot->view[vi]->box == COORD_UNDEF) return;

      plot->view[vi]->paintapi_direct->gc_set_dashes(plot->view[vi]->paintapi_direct, plot->cursor_marker_gc, -plot->view[vi]->hline, cursor_marker_dashes, SZ(cursor_marker_dashes));
      plot->view[vi]->paintapi_direct->draw_rectangle(plot->view[vi]->paintapi_direct, plot->cursor_marker_gc, 0, plot->view[vi]->bx, plot->view[vi]->by, plot->view[vi]->box, plot->view[vi]->boy);

      plot->view[vi]->box = COORD_UNDEF;
   }
}

// delayed stuff

#define DELAY 1900

static gboolean delay_callback(my_plot_t *plot) {
   plot->delayed_add_zoom_timer_tag = 0;

   // TODO maybe visual feedback
   add_to_zoom_history(plot);
   return 0;
}

void delayed_add_to_zoom_history(my_plot_t *plot) {
   if(plot->disable_delayed_add_zoom > 0) {
      //pprintf("ign\n");
      return;
   }

   pprintf("del\n");

   cancel_delayed_add_to_zoom_history(plot);

   plot->delayed_add_zoom_timer_tag = g_timeout_add(DELAY, (GSourceFunc)delay_callback, plot);
}

boolean_t commit_delayed_add_to_zoom_history(my_plot_t *plot) {
   pprintf("com\n");

   if(cancel_delayed_add_to_zoom_history(plot)) {
      add_to_zoom_history(plot);
      return true;
   }
   return false;
}

boolean_t cancel_delayed_add_to_zoom_history(my_plot_t *plot) {
   if(plot->delayed_add_zoom_timer_tag) {
      pprintf("can\n");

      g_source_remove(plot->delayed_add_zoom_timer_tag);
      plot->delayed_add_zoom_timer_tag = 0;
      return true;
   }

   return false;
}

/*
 * Local variables:
 * c-file-style: "ellemtel"
 * c-file-offsets: ((c . c-lineup-dont-change) (statement-cont . (lambda (le) (if (save-excursion (goto-char (cdr le)) (looking-at "return")) (c-lineup-java-inher le) (c-lineup-math le)))))
 * End:
 */
