/*
 * Plot - plot time-based data on screen and to file with interactive controls
 * Copyright (C) 2006  Jonas Berlin <xkr47@outerspace.dyndns.org>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 ****
 * plot2 - most functions related to representing & managing the
 * actual graphs along with gtk event handlers - except for the
 * plotting itself, see plot_draw, and zoom history, see
 * plot2-zoom-history
 */

#include <gtk/gtkbox.h>
#include <gtk/gtktogglebutton.h>
#include <math.h>

#include "../system.xh"
#include "compat.xh"
#include "eventhandlers.xh"
#include "paintapi-gtk.xh"
#include "plot2-zoom-history.xh"
#include "plot2.xh"
#include "tail_fam.xh"

#if 0
#define pprintf printf
#else
#define pprintf(...)
#endif

void reset_gc(my_plot_t *plot) {
   plot->view[0]->paintapi_direct->gc_free(plot->view[0]->paintapi_direct, plot->cursor_marker_gc);
   plot->cursor_marker_gc = NULL;

   for(int vi=0; vi<plot->view_count; ++vi) {
      plot_draw_reset_gcs(&plot->view[vi]->draw);
   }
}

void plot_init(my_plot_t *plot) {
   bzero(plot, sizeof(*plot));

   // init default view
   plot->view_dflt.parent = plot; // back ref
   plot->view_dflt.hline = -1;
   plot->view_dflt.vline = -1;
   plot->view_dflt.box = COORD_UNDEF;
   plot_draw_init(&plot->view_dflt.draw);
}

void plot_free(my_plot_t *plot) {
   for(int vi=0; vi<plot->view_count; ++vi) {
      if(plot->view[vi]->paintapi_direct) {
	 plot->view[vi]->paintapi_direct->api_free(plot->view[vi]->paintapi_direct);
      }
      plot_draw_deinit(&plot->view[vi]->draw);
      if(plot->view[vi]->draw.legend.font) {
	 free(plot->view[vi]->draw.legend.font);
      }
   }

   data_scrollbar_controller_free(plot->dsc_x);
   for(int vi=0; vi<plot->view_count; ++vi) {
      data_scrollbar_controller_free(plot->view[0]->dsc_y);
   }

   for(int vi=0; vi<plot->view_count; ++vi) {
      if(plot->view[vi]->pangofont) {
	 pango_font_description_free(plot->view[vi]->pangofont);
      }
      if(plot->view[vi]->pangoctx && !plot->view[vi]->da) {
	 g_object_unref(plot->view[vi]->pangoctx);
      }
   }

#ifdef USE_FAM
   if(plot->fam_initialized) {
      g_source_remove(plot->fam_evtsrcid);
      FAMClose(&plot->fc);
   }
#endif

   bzero(plot, sizeof(*plot));
   free(plot);
}

void plot_view_clone(my_plot_view_t *n, my_plot_view_t *o) {
   data_scrollbar_controller_copy(n->dsc_y, o->dsc_y);
   plot_draw_clone(&n->draw, &o->draw);
}

my_plot_t *plot_clone(my_plot_t *o) {
   my_plot_t *n = malloc_c(sizeof(my_plot_t));
   plot_init(n);

   my_plot_view_arr_set_length(n->view, o->view_count);

   n->dsc_x = data_scrollbar_controller_clone(o->dsc_x);
   n->fontsize_at_width = o->fontsize_at_width;
   n->scale_font_by_size = o->scale_font_by_size;
   n->cursor_marker_color = o->cursor_marker_color;

   for(int vi=0; vi<o->view_count; ++vi) {
      n->view[vi]->dsc_y = data_scrollbar_controller_clone(o->view[vi]->dsc_y);
      plot_view_clone(n->view[vi], o->view[vi]);
   }

   return n;
}

static my_plot_view_t *plot_get_prio_view(my_plot_t *plot) {
   for(int vi=0; vi<plot->view_count; ++vi) {
      if(plot->view[vi]->bmode != BMODE_OFF) {
	 return plot->view[vi];
      }
   }
   return NULL;
}

void plot_need_redraw(my_plot_t *plot) {
   //printf("need_redraw\n");
   my_plot_view_t *prio = plot_get_prio_view(plot);
   if(prio) {
      // some view is prioritized, render it and keep the others waiting
      plot_view_need_redraw(prio);
      for(int vi=0; vi<plot->view_count; ++vi) {
	 if(plot->view[vi] == prio) continue;
	 plot->view[vi]->normal_expose++;
      }
   } else {
      for(int vi=0; vi<plot->view_count; ++vi) {
	 plot_view_need_redraw(plot->view[vi]);
      }
   }
}

void plot_view_need_redraw(my_plot_view_t *view) {
   //printf("view_need_redraw\n");
   if(view->conc == 0) {
      if(view->da) gtk_widget_queue_draw(view->da);
   } else {
      view->normal_expose++;
   }
}

void plot_size_or_font_changed(my_plot_t *plot) {
   int w,h;
   for(int vi=0; vi<plot->view_count; ++vi) {
      my_plot_view_t *view = plot->view[vi];

      gdk_drawable_get_size(view->drawable, &w, &h);

      PangoFontDescription *font_desc;
      if(plot->scale_font_by_size) {
	 int fontlen = strlen(plot->font);
	 char *font = malloc((size_t)(fontlen + 40));
	 char *space = strrchr(plot->font, ' ');
	 double size;
	 if(!space) {
	    space = font + fontlen;
	    size = 12;
	 } else {
	    size = atof(space+1);
	 }
	 memcpy(font, plot->font, (size_t)(space - plot->font));

	 // scrollbars etc take 72x110 pixels
	 // good fontsizes:
	 // at 640x480, wh = 530.66 native-fontsize = 9 scaled-fontsize=11
	 // at 640x480, wh = 640 native-fontsize = 10
	 // at 1600x1200, wh = 1491.33 native-fontsize = 17
	 // at 1600x1200, wh = 1600 native-fontsize = 18
	 // first scale according to screen size

	 // scaled = 2.4^(x/1600-1) * 18.3
	 // native = 3.25^(x/1600-1) * 18.3

	 // 18.3 * pow(1.4,sw/1600-1) * pow(2.15, ((pw - 72 + (pw*3/4-110)*4/3)/2)/1600 -1 )

	 //double x1 = size;

	 int cur_screen_width = plot->view[0]->da ? gdk_screen_get_width(gtk_window_get_screen(plot->parent)) : 1600;

	 size *= pow(1.4, ((double)cur_screen_width) / 1600 - 1);

	 //double x2 = size;

	 // then scale according to widget size
	 double wh = (w + 4.0 * h / 3.0) / 2.0;

	 //size *= exp2(wh / cur_screen_width - 1);
	 //size *= exp2(wh / plot->fontsize_at_width - 1);
	 size *= pow(2.15, wh / 1600 - 1);

	 if (plot->double_font_size) {
	    size *= 2;
	 }

	 snprintf(font + (space-plot->font), (size_t)40, " %.1lf", round(size * 2) / 2);

	 //printf("font %s / %lf -> %lf -> %lf   wh %lf  %dx%d   (%d)\n", font, x1, x2, size, wh, w, h, cur_screen_width);

	 font_desc = pango_font_description_from_string(font);

	 free(font);

      } else {
	 font_desc = pango_font_description_from_string(plot->font);
      }

      if(view->pangofont) {
	 pango_font_description_free(view->pangofont);
      }

      view->pangofont = font_desc;

      if(!view->pangoctx) {
	 if(view->da) {
	    /* TODO If you create and keep a PangoLayout using this
               context, you must deal with changes to the context by
               calling pango_layout_context_changed() on the layout in
               response to the ::style-set and ::direction-changed
               signals for the widget.
             */
	    view->pangoctx = gtk_widget_get_pango_context(view->da);
	 } else {
	    view->pangoctx = gdk_pango_context_get();
	 }
      }

      // update plot_draw
      if(!view->draw.legend.font) {
	 view->draw.legend.font = zmalloc_c(sizeof(paintapi_font_t));
	 view->draw.grid.font_xaxis = view->draw.legend.font;
	 view->draw.grid.font_yaxis = view->draw.legend.font;
      }
      view->draw.legend.font->ctx = view->pangoctx;
      view->draw.legend.font->font = view->pangofont;

      view->draw.plot_xe = w - 1;
      view->draw.plot_ye = h - 1;
   }

   plot_need_redraw(plot);
}


void show_all_nohistory(my_plot_view_t *view) {
   double min, max;
   
   ++view->parent->disable_delayed_add_zoom;
   
   data_scrollbar_controller_get_range(view->parent->dsc_x, &min, &max);
   data_scrollbar_controller_set_page_length(view->parent->dsc_x, max - min);

   data_scrollbar_controller_get_range(view->dsc_y, &min, &max);
   data_scrollbar_controller_set_page_length(view->dsc_y, max - min);

   --view->parent->disable_delayed_add_zoom;

   update_cursor(view);
}

void show_all(my_plot_view_t *view) {
   show_all_nohistory(view);
   add_to_zoom_history(view->parent);
}

void pan_right(my_plot_t *plot) {
   if(gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(plot->checkbutton_tail_move))) {
      // move
      double xlength = data_scrollbar_controller_get_page_length(plot->dsc_x);

      double min, max;
      data_scrollbar_controller_get_range(plot->dsc_x, &min, &max);

      ++plot->disable_delayed_add_zoom;
   
      data_scrollbar_controller_set_page_off(plot->dsc_x, max - xlength);

      --plot->disable_delayed_add_zoom;

   } else {
      // zoom out
      double xoff = data_scrollbar_controller_get_page_off(plot->dsc_x);

      double min, max;
      data_scrollbar_controller_get_range(plot->dsc_x, &min, &max);

      ++plot->disable_delayed_add_zoom;
   
      data_scrollbar_controller_set_page_length(plot->dsc_x, max - xoff);

      --plot->disable_delayed_add_zoom;
   }
}

int plot_add_view_with_xml(my_plot_t *plot, GladeXML *gladexml) {
   int idx = plot->view_count;
   my_plot_view_arr_set_length(plot->view, idx+1);
   my_plot_view_t *view = plot->view[idx];

   view->da = glade_xml_get_widget(gladexml, "da_plot");
   gtk_widget_set_double_buffered(view->da, false);

   view->dsc_y = data_scrollbar_controller_new(gladexml, "yscroll_", 0);
   view->box_yscrolls_plot = glade_xml_get_widget(gladexml, "box_yscrolls_plot");;

   plot_view_connect_signals(view, gladexml);

   return idx;
}

void plot_add_view(my_plot_t *plot) {
   GladeXML *gladexml = glade_xml_new(plot->gladefile, "box_yscrolls_plot", NULL);
   GtkWidget *box = glade_xml_get_widget(gladexml, "box_yscrolls_plot");
   gtk_box_pack_start(GTK_BOX(plot->box_plots), box, 1, 1, 0);

   int idx = plot_add_view_with_xml(plot, gladexml);

   plot_view_clone(plot->view[idx], plot->view[idx-1]);

   if(plot->view[idx]->draw.bookmarks.show == BOOKMARKS_SHOW) {
      plot->view[idx]->draw.bookmarks.show = BOOKMARKS_SHOW_POLE_ONLY;
   }

   update_view_minmax(plot->view[idx]);
}

void plot_move_to(my_plot_view_t *view, int x, int y) {

/*
  data_scrollbar_controller_set_zoompoint(view->dsc_x, ((double)x) / view->draw.plot_xe);
  data_scrollbar_controller_set_zoompoint(view->dsc_y, 1 - ((double)y) / view->draw.plot_yde);
*/

   int xd = x - view->bx;
   int yd = view->by - y;

   xd *= view->bspeed;
   yd *= view->bspeed;

   //printf("move %d,%d\n", xd, yd);

   double xmin, xmax;
   double ymin, ymax;

   data_scrollbar_controller_get_range(view->parent->dsc_x, &xmin, &xmax);
   data_scrollbar_controller_get_range(view->dsc_y, &ymin, &ymax);

   double xlength = data_scrollbar_controller_get_page_length(view->parent->dsc_x);
   double ylength = data_scrollbar_controller_get_page_length(view->dsc_y);

   double nxoff = view->bxoff - xd * xlength / view->draw.plot_xe;
   double nyoff = view->byoff - yd * ylength / view->draw.plot_yde;

   //printf("nt %lf %lf\n", nxoff, nyoff);

   if(nxoff < xmin) nxoff = xmin; else if(nxoff + xlength > xmax) nxoff = xmax - xlength;
   if(nyoff < ymin) nyoff = ymin; else if(nyoff + ylength > ymax) nyoff = ymax - ylength;

   //printf("n %lf %lf\n", nxoff, nyoff);

   ++view->parent->disable_delayed_add_zoom;
   
   data_scrollbar_controller_set_page_off(view->parent->dsc_x, nxoff);
   data_scrollbar_controller_set_page_off(view->dsc_y, nyoff);

   --view->parent->disable_delayed_add_zoom;
}


void update_view_minmax(my_plot_view_t *view) {
   double vmin, vmax;
   double vdiffmin;

   int count = plot_draw_calc_y_minmax(&view->draw, &vmin, &vmax, &vdiffmin, true);

   // don't do anything if no lines visible
   if(count == 0) return;

   ++view->parent->disable_delayed_add_zoom;

   data_scrollbar_controller_set_range(view->dsc_y, vmin - (vmax-vmin)/20, vmax + (vmax-vmin)/20);

   //data_scrollbar_controller_set_page_length(plot->dsc_y, vmax-vmin); // initially zoom out

   data_scrollbar_controller_set_min_interval(view->dsc_y, vdiffmin * 10);

   --view->parent->disable_delayed_add_zoom;
}

void update_x_minmax(my_plot_t *plot) {
   // don't do anything if no datasets available
   if(!plot->view[0]->draw.line_set_count) return;

   my_time_t ts, te;
   my_time_t min_time_diff;

   plot_draw_calc_x_minmax(&plot->view[0]->draw, &ts, &te, &min_time_diff);

   /*
   // extended zoom out
   ts -= (te - ts) * 10;
   te += (te - ts);
   */

   ++plot->disable_delayed_add_zoom;

   data_scrollbar_controller_set_range(plot->dsc_x, (double)ts, (double)te);
   data_scrollbar_controller_set_min_interval(plot->dsc_x, min_time_diff * 10.0);

   --plot->disable_delayed_add_zoom;
}

void update_minmax(my_plot_t *plot) {
   update_x_minmax(plot);

   for(int vi=0; vi<plot->view_count; ++vi) {
      update_view_minmax(plot->view[vi]);

      update_cursor(plot->view[vi]);
   }
}


void update_cursor(my_plot_view_t *view) {
   GdkCursorType cursor = -2; // -2 = system default

   switch(view->bmode) {
      case BMODE_OFF:
	 if(view->by > view->draw.plot_yde) {
	    cursor = GDK_SB_UP_ARROW;
	 } else if(view->bbookmark_id != -1) {
	    cursor = GDK_SB_H_DOUBLE_ARROW;
	 } else if(view->inside_legend) {
	    cursor = GDK_HAND2;
	 } else if(view->parent->ctrl_pressed) {
	    cursor = GDK_TCROSS;
	 } else if(GTK_WIDGET_SENSITIVE(view->parent->button_show_all)) {
	    cursor = GDK_FLEUR;
	 } else {
	    // system default
	 }
	 break;

      case BMODE_INTERACTIVE_MOVE:
	 cursor = GDK_FLEUR;
	 break;

      case BMODE_INTERACTIVE_ZOOM:
      case BMODE_MOVE_BOOKMARK_INTERACTIVE_ZOOM:
	 cursor = GDK_SIZING;
	 break;

      case BMODE_ZOOM_AREA:
	 cursor = GDK_TCROSS;
	 break;

      case BMODE_LEGEND_ENABLE_DISABLE:
	 cursor = GDK_SIZING;
	 break;

      case BMODE_MOVE_BOOKMARK:
	 cursor = GDK_SB_UP_ARROW;
	 break;
   }

   pprintf("mode %d %s%s%s-> %d\n",
	   view->bmode,
	   view->inside_legend ? "IL " : "",
	   view->parent->ctrl_pressed ? "CP " : "",
	   GTK_WIDGET_SENSITIVE(view->parent->button_show_all) ? "WS " : "",
	   cursor);

   if(cursor != view->cursor) {
      GdkCursor *pcursor = cursor == -2 ? NULL : gdk_cursor_new(cursor);
      gdk_window_set_cursor(view->da->window, pcursor);
      if(pcursor) gdk_cursor_unref(pcursor);
      view->cursor = cursor;

      /*
      // invisible cursor code

      gchar bits[] = { 0 };
      GdkColor color = { 0, 0, 0, 0 };
      GdkCursor *cursor;

      GdkPixmap *pixmap = gdk_bitmap_create_from_data(NULL, bits, 1, 1);
      GdkCursor *cursor = gdk_cursor_new_from_pixmap(pixmap, pixmap, &color, &color, 0, 0);
      gdk_window_set_cursor(view->da->window, cursor);
      */
   }
}

/*
 * Local variables:
 * c-file-style: "ellemtel"
 * c-file-offsets: ((c . c-lineup-dont-change) (statement-cont . (lambda (le) (if (save-excursion (goto-char (cdr le)) (looking-at "return")) (c-lineup-java-inher le) (c-lineup-math le)))))
 * End:
 */
